#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from aiounittest import async_test
from swimai.recon._parsers import _ReconUtils, _OutputMessage, _InputMessage
from test.utils import CustomString


class TestUtils(unittest.TestCase):

    @async_test
    async def test_is_valid_ident_start_char_A(self):
        # Given
        character = 'A'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_start_char_Z(self):
        # Given
        character = 'Z'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_start_char_P(self):
        # Given
        character = 'P'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_start_char_a(self):
        # Given
        character = 'a'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_start_char_z(self):
        # Given
        character = 'z'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_start_char_h(self):
        # Given
        character = 'h'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_start_char_underscore(self):
        # Given
        character = '_'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_invalid_ident_start_char_digit(self):
        # Given
        character = '4'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_start_char_hyphen(self):
        # Given
        character = '-'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_start_char_dollar(self):
        # Given
        character = '$'
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_start_char_empty(self):
        # Given
        character = ''
        # When
        actual = await _ReconUtils._is_ident_start_char(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_valid_ident_char_A(self):
        # Given
        character = 'A'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_Z(self):
        # Given
        character = 'Z'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_N(self):
        # Given
        character = 'N'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_a(self):
        # Given
        character = 'a'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_z(self):
        # Given
        character = 'z'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_g(self):
        # Given
        character = 'g'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_underscore(self):
        # Given
        character = '_'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_hyphen(self):
        # Given
        character = '-'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_0(self):
        # Given
        character = '0'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_9(self):
        # Given
        character = '9'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_ident_char_5(self):
        # Given
        character = '5'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_invalid_ident_char_dollar(self):
        # Given
        character = '$'
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_char_empty(self):
        # Given
        character = ''
        # When
        actual = await _ReconUtils._is_ident_char(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_valid_ident(self):
        # Given
        value = 'test1'
        # When
        actual = await _ReconUtils._is_ident(value)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_invalid_ident_empty(self):
        # Given
        value = ''
        # When
        actual = await _ReconUtils._is_ident(value)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_invalid_start_char(self):
        # Given
        value = '1test'
        # When
        actual = await _ReconUtils._is_ident(value)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_invalid_middle_char(self):
        # Given
        value = 'test$test'
        # When
        actual = await _ReconUtils._is_ident(value)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_invalid_end_char(self):
        # Given
        value = 'test$'
        # When
        actual = await _ReconUtils._is_ident(value)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_ident_invalid_all_chars(self):
        # Given
        value = '^%&*'
        # When
        actual = await _ReconUtils._is_ident(value)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_valid_space_char_space(self):
        # Given
        character = ' '
        # When
        actual = await _ReconUtils._is_space(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_space_char_tab(self):
        # Given
        character = '\t'
        # When
        actual = await _ReconUtils._is_space(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_invalid_space_char_digit(self):
        # Given
        character = '5'
        # When
        actual = await _ReconUtils._is_space(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_space_char_letter(self):
        # Given
        character = 'l'
        # When
        actual = await _ReconUtils._is_space(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_space_char_empty(self):
        # Given
        character = ''
        # When
        actual = await _ReconUtils._is_space(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_valid_digit_char_1(self):
        # Given
        character = '2'
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_digit_char_0(self):
        # Given
        character = '0'
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_digit_char_7(self):
        # Given
        character = '7'
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_valid_digit_char_9(self):
        # Given
        character = '9'
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_is_invalid_digit_char_letter(self):
        # Given
        character = 'b'
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_digit_char_underscore(self):
        # Given
        character = '_'
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_is_invalid_digit_char_empty(self):
        # Given
        character = ''
        # When
        actual = await _ReconUtils._is_digit(character)
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_to_ord_str(self):
        # Given
        character = 'p'
        # When
        actual = await _ReconUtils._to_ord(character)
        # Then
        self.assertEqual(112, actual)

    @async_test
    async def test_to_ord_int(self):
        # Given
        character = 3
        # When
        actual = await _ReconUtils._to_ord(character)
        # Then
        self.assertEqual(3, actual)

    @async_test
    async def test_to_ord_other(self):
        # Given
        character = 3.12
        # When
        actual = await _ReconUtils._to_ord(character)
        # Then
        self.assertEqual(None, actual)

    @async_test
    async def test_output_message_create_empty(self):
        # Given
        chars = None
        # When
        actual = await _OutputMessage._create(chars)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('', actual._value)
        self.assertEqual(0, actual._size)
        self.assertEqual('', actual._last_char)

    @async_test
    async def test_output_message_create_single(self):
        # Given
        chars = 'p'
        # When
        actual = await _OutputMessage._create(chars)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('p', actual._value)
        self.assertEqual(1, actual._size)
        self.assertEqual('p', actual._last_char)

    @async_test
    async def test_output_message_create_multiple(self):
        # Given
        chars = 'foo_bar'
        # When
        actual = await _OutputMessage._create(chars)
        # Then
        self.assertIsInstance(actual, _OutputMessage)
        self.assertEqual('foo_bar', actual._value)
        self.assertEqual(7, actual._size)
        self.assertEqual('r', actual._last_char)

    @async_test
    async def test_output_message_append_str_to_empty(self):
        # Given
        output_message = await _OutputMessage._create('')
        chars = 'bar'
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('bar', output_message._value)
        self.assertEqual(3, output_message._size)
        self.assertEqual('r', output_message._last_char)

    @async_test
    async def test_output_message_append_float_to_empty(self):
        # Given
        output_message = await _OutputMessage._create('')
        chars = 21.12
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('21.12', output_message._value)
        self.assertEqual(5, output_message._size)
        self.assertEqual('2', output_message._last_char)

    @async_test
    async def test_output_message_append_int_to_empty(self):
        # Given
        output_message = await _OutputMessage._create('')
        chars = 13
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('13', output_message._value)
        self.assertEqual(2, output_message._size)
        self.assertEqual('3', output_message._last_char)

    @async_test
    async def test_output_message_append_output_message_to_empty(self):
        # Given
        output_message = await _OutputMessage._create('')
        chars = await _OutputMessage._create('boo_fast')
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('boo_fast', output_message._value)
        self.assertEqual(8, output_message._size)
        self.assertEqual('t', output_message._last_char)

    @async_test
    async def test_output_message_append_input_message_to_empty(self):
        # Given
        output_message = await _OutputMessage._create('')
        chars = await _InputMessage._create('my_message')
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('my_message', output_message._value)
        self.assertEqual(10, output_message._size)
        self.assertEqual('e', output_message._last_char)

    @async_test
    async def test_output_message_append_invalid_to_empty(self):
        # Given
        output_message = await _OutputMessage._create('')
        chars = CustomString('moo')
        # When
        with self.assertRaises(TypeError) as error:
            await output_message._append(chars)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Item of type CustomString cannot be added to Message!', message)

    @async_test
    async def test_output_message_append_str_to_existing(self):
        # Given
        output_message = await _OutputMessage._create('full')
        chars = '_qux'
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('full_qux', output_message._value)
        self.assertEqual(8, output_message._size)
        self.assertEqual('x', output_message._last_char)

    @async_test
    async def test_output_message_append_float_to_existing(self):
        # Given
        output_message = await _OutputMessage._create('full')
        chars = 22.1
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('full22.1', output_message._value)
        self.assertEqual(8, output_message._size)
        self.assertEqual('1', output_message._last_char)

    @async_test
    async def test_output_message_append_int_to_existing(self):
        # Given
        output_message = await _OutputMessage._create('empty')
        chars = 73
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('empty73', output_message._value)
        self.assertEqual(7, output_message._size)
        self.assertEqual('3', output_message._last_char)

    @async_test
    async def test_output_message_append_output_message_to_existing(self):
        # Given
        output_message = await _OutputMessage._create('empty')
        chars = await _OutputMessage._create(' house')
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('empty house', output_message._value)
        self.assertEqual(11, output_message._size)
        self.assertEqual('e', output_message._last_char)

    @async_test
    async def test_output_message_append_input_message_to_existing(self):
        # Given
        output_message = await _OutputMessage._create('input_')
        chars = await _InputMessage._create('message')
        # When
        await output_message._append(chars)
        # Then
        self.assertIsInstance(output_message, _OutputMessage)
        self.assertEqual('input_message', output_message._value)
        self.assertEqual(13, output_message._size)
        self.assertEqual('e', output_message._last_char)

    @async_test
    async def test_output_message_append_invalid_to_existing(self):
        # Given
        output_message = await _OutputMessage._create('bar')
        chars = CustomString('foo')
        # When
        with self.assertRaises(TypeError) as error:
            await output_message._append(chars)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Item of type CustomString cannot be added to Message!', message)

    @async_test
    async def test_input_message_create_empty(self):
        # Given
        chars = None
        # When
        actual = await _InputMessage._create(chars)
        # Then
        self.assertIsInstance(actual, _InputMessage)
        self.assertEqual('', actual._value)
        self.assertEqual(0, actual._size)
        self.assertEqual('', actual._head)
        self.assertFalse(actual._is_cont)

    @async_test
    async def test_input_message_create_single(self):
        # Given
        chars = 'm'
        # When
        actual = await _InputMessage._create(chars)
        # Then
        self.assertIsInstance(actual, _InputMessage)
        self.assertEqual('m', actual._value)
        self.assertEqual(1, actual._size)
        self.assertEqual('m', actual._head)
        self.assertTrue(actual._is_cont)

    @async_test
    async def test_input_message_create_multiple(self):
        # Given
        chars = 'moo_cow'
        # When
        actual = await _InputMessage._create(chars)
        # Then
        self.assertIsInstance(actual, _InputMessage)
        self.assertEqual('moo_cow', actual._value)
        self.assertEqual(7, actual._size)
        self.assertEqual('m', actual._head)
        self.assertTrue(actual._is_cont)

    @async_test
    async def test_input_message_append_str_to_empty(self):
        # Given
        input_message = await _InputMessage._create('')
        chars = 'test'
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('test', input_message._value)
        self.assertEqual(4, input_message._size)
        self.assertEqual('t', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_float_to_empty(self):
        # Given
        input_message = await _InputMessage._create('')
        chars = -0.12
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('-0.12', input_message._value)
        self.assertEqual(5, input_message._size)
        self.assertEqual('-', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_int_to_empty(self):
        # Given
        input_message = await _InputMessage._create('')
        chars = -32
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('-32', input_message._value)
        self.assertEqual(3, input_message._size)
        self.assertEqual('-', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_input_message_to_empty(self):
        # Given
        input_message = await _InputMessage._create('')
        chars = await _InputMessage._create('input_bar')
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('input_bar', input_message._value)
        self.assertEqual(9, input_message._size)
        self.assertEqual('i', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_output_message_to_empty(self):
        # Given
        input_message = await _InputMessage._create('')
        chars = await _OutputMessage._create('output_bar')
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('output_bar', input_message._value)
        self.assertEqual(10, input_message._size)
        self.assertEqual('o', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_invalid_to_empty(self):
        # Given
        input_message = await _InputMessage._create('')
        chars = None
        # When
        with self.assertRaises(TypeError) as error:
            await input_message._append(chars)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Item of type NoneType cannot be added to Message!', message)

    @async_test
    async def test_input_message_append_str_to_existing(self):
        # Given
        input_message = await _InputMessage._create('cow')
        chars = '_moo'
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('cow_moo', input_message._value)
        self.assertEqual(7, input_message._size)
        self.assertEqual('c', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_float_to_existing(self):
        # Given
        input_message = await _InputMessage._create('paw_')
        chars = -1.23
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('paw_-1.23', input_message._value)
        self.assertEqual(9, input_message._size)
        self.assertEqual('p', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_int_to_existing(self):
        # Given
        input_message = await _InputMessage._create('+-111')
        chars = -1112
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('+-111-1112', input_message._value)
        self.assertEqual(10, input_message._size)
        self.assertEqual('+', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_input_message_to_existing(self):
        # Given
        input_message = await _InputMessage._create('+_/2')
        chars = await _InputMessage._create('10-1-2')
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('+_/210-1-2', input_message._value)
        self.assertEqual(10, input_message._size)
        self.assertEqual('+', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_output_message_to_existing(self):
        # Given
        input_message = await _InputMessage._create('bar')
        chars = await _OutputMessage._create('_foo')
        # When
        await input_message._append(chars)
        # Then
        self.assertIsInstance(input_message, _InputMessage)
        self.assertEqual('bar_foo', input_message._value)
        self.assertEqual(7, input_message._size)
        self.assertEqual('b', input_message._head)
        self.assertTrue(input_message._is_cont)

    @async_test
    async def test_input_message_append_invalid_to_existing(self):
        # Given
        input_message = await _InputMessage._create('bar')
        chars = None
        # When
        with self.assertRaises(TypeError) as error:
            await input_message._append(chars)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Item of type NoneType cannot be added to Message!', message)

    @async_test
    async def test_input_message_skip_spaces_none(self):
        # Given
        message = await _InputMessage._create('foo')
        # When
        await message._skip_spaces(message)
        # Then
        self.assertEqual('f', message._head)

    @async_test
    async def test_input_message_skip_spaces_single(self):
        # Given
        message = await _InputMessage._create(' bar')
        # When
        await message._skip_spaces(message)
        # Then
        self.assertEqual('b', message._head)

    @async_test
    async def test_input_message_skip_spaces_multiple(self):
        # Given
        message = await _InputMessage._create('    qux')
        # When
        await message._skip_spaces(message)
        # Then
        self.assertEqual('q', message._head)

    @async_test
    async def test_input_message_skip_spaces_spaces_only(self):
        # Given
        message = await _InputMessage._create('    ')
        # When
        await message._skip_spaces(message)
        # Then
        self.assertEqual('', message._head)

    @async_test
    async def test_input_message_head_empty(self):
        # Given
        message = await _InputMessage._create('')
        # When
        actual = message._head
        # Then
        self.assertEqual('', actual)

    @async_test
    async def test_input_message_head_continuous(self):
        # Given
        message = await _InputMessage._create('abc')
        # When
        message._step()
        actual = message._head
        # Then
        self.assertEqual('b', actual)

    @async_test
    async def test_input_message_head_not_continuous(self):
        # Given
        message = await _InputMessage._create('abc')
        # When
        message._step()
        message._step()
        message._step()
        actual = message._head
        # Then
        self.assertEqual('', actual)

    @async_test
    async def test_input_message_is_continuous_empty(self):
        # Given
        message = await _InputMessage._create('')
        # When
        actual = message._is_cont
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_input_message_is_continuous_existing_true(self):
        # Given
        message = await _InputMessage._create('dog')
        # When
        actual = message._is_cont
        # Then
        self.assertTrue(actual)

    @async_test
    async def test_input_message_is_continuous_existing_false(self):
        # Given
        message = await _InputMessage._create('dog')
        # When
        message._step()
        message._step()
        message._step()
        actual = message._is_cont
        # Then
        self.assertFalse(actual)

    @async_test
    async def test_input_message_step_once(self):
        # Given
        message = await _InputMessage._create('dog')
        # When
        actual = message._step()
        # Then
        self.assertEqual('o', actual)

    @async_test
    async def test_input_message_step_twice(self):
        # Given
        message = await _InputMessage._create('dog')
        # When
        message._step()
        actual = message._step()
        # Then
        self.assertEqual('g', actual)

    @async_test
    async def test_input_message_step_out_of_bound(self):
        # Given
        message = await _InputMessage._create('dog')
        # When
        message._step()
        message._step()
        message._step()
        actual = message._step()
        # Then
        self.assertEqual('', actual)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from urllib.parse import urlparse

from swimai.client._utils import _URI


class TestUtils(unittest.TestCase):

    def test_uri_scheme_valid_ws(self):
        # Given
        uri = 'ws://foo_bar:9000'
        uri = urlparse(uri)
        # When
        actual = _URI._has_valid_scheme(uri)
        # Then
        self.assertTrue(actual)

    def test_uri_scheme_valid_warp(self):
        # Given
        uri = 'warp://foo_bar:9000'
        uri = urlparse(uri)
        # When
        actual = _URI._has_valid_scheme(uri)
        # Then
        self.assertTrue(actual)

    def test_uri_scheme_invalid_empty(self):
        # Given
        uri = 'foo_bar:9000'
        uri = urlparse(uri)
        # When
        actual = _URI._has_valid_scheme(uri)
        # Then
        self.assertFalse(actual)

    def test_uri_scheme_invalid_http(self):
        # Given
        uri = 'carp://foo_bar:9000'
        uri = urlparse(uri)
        # When
        actual = _URI._has_valid_scheme(uri)
        # Then
        self.assertFalse(actual)

    def test_normalise_warp_scheme_ws(self):
        # Given
        uri = 'ws://foo_bar:9000'
        expected = 'ws://foo_bar:9000'
        # When
        actual = _URI._normalise_warp_scheme(uri)
        # Then
        self.assertEqual(expected, actual)

    def test_normalise_warp_scheme_warp(self):
        # Given
        uri = 'warp://foo_bar:9000'
        expected = 'ws://foo_bar:9000'
        # When
        actual = _URI._normalise_warp_scheme(uri)
        # Then
        self.assertEqual(expected, actual)

    def test_normalise_warp_scheme_invalid(self):
        # Given
        uri = 'carp://foo_bar:9000'
        # When
        with self.assertRaises(TypeError) as error:
            _URI._normalise_warp_scheme(uri)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Invalid scheme "carp" for Warp URI!', message)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import unittest

from swimai.structures import Num, Attr, Slot, Text, RecordMap, Bool, Value, RecordConverter
from swimai.structures._structs import _Item, _Extant, _Absent, _RecordFlags, _Record, _RecordMapView, _ValueBuilder
from test.utils import CustomString, CustomItem, MockPerson, MockPet, MockCar


class TestStructs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        file_location = os.path.dirname(__file__)
        with open(os.path.join(file_location, 'expected_strings.txt')) as golden_file:
            cls.expected_strings = dict()

            for line in golden_file:
                (key, val) = line.strip().split('=')
                cls.expected_strings[key] = val

    def get_name(self):
        return self.id().split('.')[-1]

    def test_item_concat_single_num(self):
        # Given
        headers = _Record.create()._add_slot('node', 'foo')._add_slot('lane', 'bar')
        body = Num.create_from(42)
        # When
        actual = Attr.create_attr('tag', headers)._concat(body)
        # Then
        self.assertEqual(2, actual.size)
        self.assertIsInstance(actual.get_items()[0], Attr)
        self.assertIsInstance(actual.get_items()[1], Num)
        self.assertEqual(42, actual.get_items()[1].value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_single_record(self):
        # Given
        headers = _Record.create()._add_slot('node', 'foo')._add_slot('lane', 'bar')
        body = _Record.create()._add_slot('attr', 'hello')._add_slot('message', 'world')
        # When
        actual = Attr.create_attr('tag', headers)._concat(body)
        # Then
        self.assertEqual(3, actual.size)
        self.assertIsInstance(actual.get_items()[0], Attr)
        self.assertIsInstance(actual.get_items()[1], Slot)
        self.assertIsInstance(actual.get_items()[2], Slot)
        self.assertEqual('attr', actual.get_items()[1].key.value)
        self.assertEqual('hello', actual.get_items()[1].value.value)
        self.assertEqual('message', actual.get_items()[2].key.value)
        self.assertEqual('world', actual.get_items()[2].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_multiple_text(self):
        # Given
        headers = _Record.create()._add_slot('node', 'foo')._add_slot('lane', 'bar')
        headers.add(Num.create_from(42))
        body = Text.create_from('Polly the parrot')
        # When
        actual = headers._concat(body)
        # Then
        self.assertEqual(2, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(3, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Text)
        self.assertEqual('Polly the parrot', actual.get_items()[1].value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_multiple_record(self):
        # Given
        headers = _Record.create()._add_slot('node', 'foo')._add_slot('lane', 'bar')
        headers.add(_Record.create()._add_slot('node', 'boo')._add_slot('lane', 'far'))
        body = _Record.create()._add_slot('node', 'poo')._add_slot('lane', 'car')
        # When
        actual = headers._concat(body)
        # Then
        self.assertEqual(3, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(3, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Slot)
        self.assertEqual('node', actual.get_items()[1].key.value)
        self.assertEqual('poo', actual.get_items()[1].value.value)
        self.assertEqual('lane', actual.get_items()[2].key.value)
        self.assertEqual('car', actual.get_items()[2].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_zero_bool(self):
        # Given
        headers = _Record.create()
        body = Bool.create_from(True)
        # When
        actual = headers._concat(body)
        # Then
        self.assertEqual(2, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(0, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Bool)
        self.assertTrue(actual.get_items()[1].value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_concat_zero_record(self):
        # Given
        headers = _Record.create()
        body = _Record.create()._add_slot('node', 'moo')._add_slot('lane', 'cow')
        # When
        actual = headers._concat(body)
        # Then
        self.assertEqual(3, actual.size)
        self.assertIsInstance(actual.get_items()[0], RecordMap)
        self.assertEqual(0, actual.get_items()[0].size)
        self.assertIsInstance(actual.get_items()[1], Slot)
        self.assertIsInstance(actual.get_items()[2], Slot)
        self.assertEqual('node', actual.get_items()[1].key.value)
        self.assertEqual('moo', actual.get_items()[1].value.value)
        self.assertEqual('lane', actual.get_items()[2].key.value)
        self.assertEqual('cow', actual.get_items()[2].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_extant(self):
        # Given
        actual = _Item.extant()
        # Then
        self.assertEqual(_Extant._get_extant(), actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_absent(self):
        # Given
        actual = _Item.absent()
        # Then
        self.assertEqual(_Absent._get_absent(), actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_create_from_item(self):
        # Given
        record = _Record.create()._add_slot('node', 'boo')._add_slot('lane', 'ghost')
        # When
        actual = _Item.create_from(record)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(2, actual.size)
        self.assertEqual('node', actual.get_items()[0].key.value)
        self.assertEqual('boo', actual.get_items()[0].value.value)
        self.assertEqual('lane', actual.get_items()[1].key.value)
        self.assertEqual('ghost', actual.get_items()[1].value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_create_from_tuple(self):
        # Given
        record = {'node': 'boo'}
        # When
        dict()
        actual = _Item.create_from(record)
        # Then
        self.assertIsInstance(actual, Slot)
        self.assertEqual('node', actual.key)
        self.assertEqual('boo', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_item_create_from_other(self):
        # Given
        record = 'monkey'
        # When
        actual = _Item.create_from(record)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('monkey', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_attr_from_text_key(self):
        # Given
        key = Text.create_from('Ghost')
        value = 'Boo'
        # When
        actual = Attr.create_attr(key, value)
        # Then
        self.assertIsInstance(actual.key, Text)
        self.assertEqual('Ghost', actual.key.value)
        self.assertEqual('Boo', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_attr_from_string_key(self):
        # Given
        key = 'Foo'
        value = Text.create_from('Bar')
        # When
        actual = Attr.create_attr(key, value)
        # Then
        self.assertIsInstance(actual.key, Text)
        self.assertEqual('Foo', actual.key.value)
        self.assertIsInstance(actual.value, Text)
        self.assertEqual('Bar', actual.value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_attr_from_invalid_key(self):
        # Given
        key = 34
        value = 'Bar'
        # When
        with self.assertRaises(TypeError) as error:
            Attr.create_attr(key, value)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Invalid key: 34', message)

    def test_create_attr_from_empty_key(self):
        # Given
        key = None
        value = 'Bar'
        # When
        with self.assertRaises(TypeError) as error:
            Attr.create_attr(key, value)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Empty key for attribute!', message)

    def test_create_attr_from_empty_value(self):
        # Given
        key = 'Foo'
        value = None
        # When
        with self.assertRaises(TypeError) as error:
            Attr.create_attr(key, value)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Empty value for attribute!', message)

    def test_create_attr_from_extant_value(self):
        # Given
        key = 'Moo'
        # When
        actual = Attr.create_attr(key, Value.extant())
        # Then
        self.assertIsInstance(actual.key, Text)
        self.assertEqual('Moo', actual.key.value)
        self.assertIsInstance(actual.value, _Extant)
        self.assertEqual(Value.extant(), actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_key_equals_str_true(self):
        # Given
        attribute = Attr.create_attr('Foo', 'Bar')
        # When
        actual = attribute.key_equals('Foo')
        # Then
        self.assertTrue(actual)

    def test_key_equals_str_false(self):
        # Given
        attribute = Attr.create_attr('Foo', 'Bar')
        # When
        actual = attribute.key_equals('Moo')
        # Then
        self.assertFalse(actual)

    def test_key_equals_attr_true(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = Attr.create_attr('Hello', 'Friend')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertTrue(actual)

    def test_key_equals_attr_false(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = Attr.create_attr('Goodbye', 'Friend')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertFalse(actual)

    def test_key_equals_other_false(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = CustomString('Hello')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertTrue(actual)

    def test_key_equals_other_true(self):
        # Given
        attribute = Attr.create_attr('Hello', 'World')
        item = CustomString('Goodbye')
        # When
        actual = attribute.key_equals(item)
        # Then
        self.assertFalse(actual)

    def test_empty_value(self):
        # Given
        actual = Value()
        # Then
        self.assertEqual(_Absent._get_absent(), actual.key)
        self.assertEqual(_Absent._get_absent(), actual.value)
        self.assertEqual(0, actual.size)

    def test_create_value_from_object_none(self):
        # Given
        obj = None
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertEqual(_Extant._get_extant(), actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_value(self):
        # Given
        obj = Text.create_from('Boom')
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Value)
        self.assertEqual('Boom', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_str(self):
        # Given
        obj = 'Dog'
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('Dog', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_int(self):
        # Given
        obj = 33
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(33, actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_float(self):
        # Given
        obj = 3.14
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(3.14, actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_boolean(self):
        # Given
        obj = True
        # When
        actual = Value.create_from(obj)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertTrue(actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_value_from_object_invalid(self):
        # Given
        obj = CustomString('Foo')
        # When
        with self.assertRaises(TypeError) as error:
            Value.create_from(obj)
        # Then
        message = error.exception.args[0]
        self.assertEqual('CustomString(Foo) cannot be converted to Value!', message)

    def test_create_text_from_string(self):
        # Given
        string = 'Tea'
        # When
        actual = Text.create_from(string)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('Tea', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_text_from_empty_string(self):
        # Given
        string = ''
        # When
        actual = Text.create_from(string)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('', actual.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_create_text_from_invalid(self):
        # Given
        string = 12
        # When
        with self.assertRaises(Exception) as error:
            # noinspection PyTypeChecker
            Text.create_from(string)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Cannot create a Text object with non string value!', message)

    def test_text_get_empty_first_time(self):
        # When
        actual = Text.get_empty()
        # Then
        self.assertEqual('', actual.value)
        self.assertEqual(Text.empty, actual)

    def test_text_get_empty_multiple_times(self):
        # Given
        Text.get_empty()
        # When
        actual = Text.get_empty()
        # Then
        self.assertEqual('', actual.value)
        self.assertEqual(Text.empty, actual)

    def test_text_string_value(self):
        # Given
        text = Text.create_from('Dog')
        # When
        actual = text.get_string_value()
        # Then
        self.assertIsInstance(actual, str)
        self.assertEqual('Dog', actual)

    def test_num_integer(self):
        # When
        num = Num(33)
        # Then
        self.assertEqual(33, num.value)
        self.assertEqual(33, num.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_num_float(self):
        # When
        num = Num(0.11)
        # Then
        self.assertEqual(0.11, num.value)
        self.assertEqual(0.11, num.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_positive_integer(self):
        # Given
        num = 99
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(99, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_negative_integer(self):
        # Given
        num = -99
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(-99, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_positive_float(self):
        # Given
        num = 9.9
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(9.9, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_negative_float(self):
        # Given
        num = -9.9
        # When
        actual = Num.create_from(num)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(-9.9, actual.get_num_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(num))

    def test_create_num_from_invalid(self):
        # Given
        num = 'False'
        # When
        with self.assertRaises(Exception) as error:
            # noinspection PyTypeChecker
            Num.create_from(num)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Cannot create a Num object with non numeric value!', message)

    def test_bool_true(self):
        # Given
        boolean = Bool(True)
        # Then
        self.assertTrue(boolean.value)
        self.assertTrue(boolean.get_bool_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(boolean))
        self.assertTrue(boolean)

    def test_bool_false(self):
        # Given
        boolean = Bool(False)
        # Then
        self.assertFalse(boolean.value)
        self.assertFalse(boolean.get_bool_value())
        self.assertEqual(self.expected_strings.get(self.get_name()), str(boolean))
        self.assertFalse(boolean)

    def test_create_bool_from_true_once(self):
        # Given
        boolean = True
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool._TRUE, actual)
        self.assertTrue(actual.get_bool_value())
        self.assertTrue(actual)

    def test_create_bool_from_true_twice(self):
        # Given
        boolean = True
        Bool.create_from(boolean)
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool._TRUE, actual)
        self.assertTrue(actual.get_bool_value())
        self.assertTrue(actual)

    def test_create_bool_from_false_once(self):
        # Given
        boolean = False
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool._FALSE, actual)
        self.assertFalse(actual.get_bool_value())
        self.assertFalse(actual)

    def test_create_bool_from_false_twice(self):
        # Given
        boolean = False
        Bool.create_from(boolean)
        # When
        actual = Bool.create_from(boolean)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool._FALSE, actual)
        self.assertFalse(actual.get_bool_value())
        self.assertFalse(actual)

    def test_get_absent_once(self):
        # When
        actual = _Absent._get_absent()
        # Then
        self.assertIsInstance(actual, _Absent)
        self.assertEqual(_Absent._absent, actual)
        self.assertEqual(_Absent._get_absent(), actual)
        self.assertFalse(_Absent._get_absent())
        self.assertFalse(actual)
        self.assertFalse(_Absent._absent)

    def test_get_absent_twice(self):
        # Given
        _Absent._get_absent()
        # When
        actual = _Absent._get_absent()
        # Then
        self.assertIsInstance(actual, _Absent)
        self.assertEqual(_Absent._absent, actual)
        self.assertEqual(_Absent._get_absent(), actual)
        self.assertFalse(actual)
        self.assertFalse(_Absent._get_absent())
        self.assertFalse(_Absent._absent)

    def test_get_extant_once(self):
        # When
        actual = _Extant._get_extant()
        # Then
        self.assertIsInstance(actual, _Extant)
        self.assertEqual(_Extant.extant, actual)
        self.assertEqual(_Extant._get_extant(), actual)
        self.assertFalse(_Extant.extant)
        self.assertFalse(actual)
        self.assertFalse(_Extant._get_extant())

    def test_get_extant_twice(self):
        # Given
        _Extant._get_extant()
        # When
        actual = _Extant._get_extant()
        # Then
        self.assertIsInstance(actual, _Extant)
        self.assertEqual(_Extant.extant, actual)
        self.assertEqual(_Extant._get_extant(), actual)
        self.assertFalse(_Extant.extant)
        self.assertFalse(actual)
        self.assertFalse(_Extant._get_extant())

    def test_slot_key_value(self):
        # Given
        slot = Slot('Foo', 'Bar')
        # Then
        self.assertIsInstance(slot, Slot)
        self.assertEqual('Foo', slot.key)
        self.assertEqual('Bar', slot.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(slot))

    def test_slot_key_only(self):
        # Given
        slot = Slot('Foo')
        # Then
        self.assertIsInstance(slot, Slot)
        self.assertEqual('Foo', slot.key)
        self.assertEqual(_Extant._get_extant(), slot.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(slot))

    def test_create_slot_key_value(self):
        # Given
        key = 'dog'
        value = 'bark'
        # When
        actual = Slot.create_slot(key, value)
        # Then
        self.assertIsInstance(actual, Slot)
        self.assertEqual('dog', actual.key)
        self.assertEqual('bark', actual.value)

    def test_create_slot_key_only(self):
        # Given
        key = 'dog'
        # When
        actual = Slot.create_slot(key)
        # Then
        self.assertIsInstance(actual, Slot)
        self.assertEqual('dog', actual.key)
        self.assertEqual(_Extant._get_extant(), actual.value)

    def test_create_slot_no_key(self):
        # Given
        key = None
        # When
        with self.assertRaises(TypeError) as error:
            Slot.create_slot(key)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Empty key for slot!', message)

    def test_record_create(self):
        # When
        actual = _Record.create()
        # Then
        self.assertIsInstance(actual, RecordMap)

    def test_record_add_all_single(self):
        # Given
        actual = _Record.create()
        item = Text.create_from('Foo')
        items = [item]
        # When
        actual_response = actual._add_all(items)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(1, actual._item_count)
        self.assertEqual(item, actual.get_item(0))
        self.assertTrue(actual_response)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_all_multiple(self):
        # Given
        actual = _Record.create()
        first_item = Text.create_from('Foo')
        second_item = Text.create_from('Bar')
        third_item = Text.create_from('Baz')
        items = [first_item, second_item, third_item]
        # When
        actual_response = actual._add_all(items)
        # Then
        self.assertEqual(3, actual.size)
        self.assertEqual(3, actual._item_count)
        self.assertEqual(first_item, actual.get_item(0))
        self.assertEqual(second_item, actual.get_item(1))
        self.assertEqual(third_item, actual.get_item(2))
        self.assertTrue(actual_response)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_all_empty(self):
        # Given
        actual = _Record.create()
        items = []
        # When
        actual_response = actual._add_all(items)
        # Then
        self.assertEqual(0, actual.size)
        self.assertEqual(0, actual._item_count)
        self.assertEqual(False, actual_response)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_item_value_item(self):
        # Given
        actual = _Record.create()
        key = Text.create_from('Good')
        value = Text.create_from('Dog')
        # When
        actual = actual._add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, _Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual(key, actual.get_item(0).key)
        self.assertEqual(value, actual.get_item(0).value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_string_value_item(self):
        # Given
        actual = _Record.create()
        key = 'Baz'
        value = Text.create_from('Qux')
        # When
        actual = actual._add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, _Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('Baz', actual.get_item(0).key.value)
        self.assertEqual('Qux', actual.get_item(0).value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_item_value_string(self):
        # Given
        actual = _Record.create()
        key = Text.create_from('Foo')
        value = 'Bar'
        # When
        actual = actual._add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, _Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('Foo', actual.get_item(0).key.value)
        self.assertEqual('Bar', actual.get_item(0).value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_add_slot_key_string_value_string(self):
        # Given
        actual = _Record.create()
        key = 'Wibble'
        value = 'Wobble'
        # When
        actual = actual._add_slot(key, value)
        # Then
        self.assertEqual(1, actual.size)
        self.assertIsInstance(actual, _Record)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertEqual('Wibble', actual.get_item(0).key.value)
        self.assertEqual('Wobble', actual.get_item(0).value.value)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_get_headers_record(self):
        # Given
        record = _Record.create()
        body = _Record.create()
        body._add_slot('Foo', 'Bar')
        record.add(Attr.create_attr(Text.create_from('cat'), body))
        # When
        actual = record._get_headers('cat')
        # Then
        self.assertIsInstance(actual, _Record)
        self.assertEqual(1, actual.size)
        self.assertEqual('Foo', actual.get_item(0).key.value)
        self.assertEqual('Bar', actual.get_item(0).value.value)

    def test_record_get_headers_object(self):
        # Given
        record = _Record.create()
        body = Slot.create_slot('Baz', 'Qux')
        record.add(Attr.create_attr(Text.create_from('dog'), body))
        # When
        actual = record._get_headers('dog')
        # Then
        self.assertIsInstance(actual, _Record)
        self.assertEqual(1, actual.size)
        self.assertEqual('Baz', actual.get_item(0).key)
        self.assertEqual('Qux', actual.get_item(0).value)

    def test_record_get_headers_none(self):
        # Given
        record = _Record.create()
        body = Slot.create_slot('Hello', 'World')
        record.add(Attr.create_attr(Text.create_from('mouse'), body))
        # When
        actual = record._get_headers('bird')
        # Then
        self.assertIsNone(actual)

    def test_record_get_head(self):
        # Given
        record = _Record.create()
        body = Slot.create_slot('Hello', 'World')
        first_item = Attr.create_attr(Text.create_from('bird'), body)
        second_item = Attr.create_attr(Text.create_from('frog'), body)
        record.add(first_item)
        record.add(second_item)
        # When
        actual = record._get_head()
        # Then
        self.assertEqual(first_item, actual)
        self.assertNotEqual(second_item, actual)

    def test_record_bind(self):
        # Given
        record = _Record.create()
        body = Slot.create_slot('Hello', 'Friend')
        first_item = Attr.create_attr(Text.create_from('dog'), body)
        record.add(first_item)
        # When
        actual = record._bind()
        # Then
        self.assertEqual(record, actual)

    def test_record_map_size_empty(self):
        # Given
        record = RecordMap.create()
        # When
        actual = record.size
        # Then
        self.assertEqual(0, actual)

    def test_record_map_size_one(self):
        # Given
        record = RecordMap.create_record_map(Text.create_from('Test'))
        # When
        actual = record.size
        # Then
        self.assertEqual(1, actual)

    def test_record_map_size_three(self):
        # Given
        record = RecordMap.create()
        record.add(Text.create_from('This'))
        record.add(Text.create_from('is'))
        record.add(Text.create_from('test'))
        # When
        actual = record.size
        # Then
        self.assertEqual(3, actual)

    def test_record_map_tag_attr(self):
        # Given
        record = RecordMap.create()
        record.add(Attr.create_attr('Foo', 'Bar'))
        # When
        actual = record._tag
        # Then
        self.assertEqual('Foo', actual)

    def test_record_map_tag_slot(self):
        # Given
        record = RecordMap.create()
        record.add(Slot.create_slot('Foo', 'Bar'))
        # When
        actual = record._tag
        # Then
        self.assertEqual(None, actual)

    def test_record_map_tag_empty(self):
        # Given
        record = RecordMap.create()
        # When
        actual = record._tag
        # Then
        self.assertEqual(None, actual)

    def test_record_map_create(self):
        # When
        actual = RecordMap.create()
        # Then
        self.assertEqual(0, actual.size)
        self.assertEqual(_RecordFlags.ALIASED.value, actual._flags)

    def test_record_map_create_with_empty(self):
        # Given
        item = None
        # When
        actual = RecordMap.create_record_map(item)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(0, actual._field_count)
        self.assertEqual(_Extant._get_extant(), actual.get_item(0))

    def test_record_map_create_with_field(self):
        # Given
        item = Attr.create_attr('Foo', 'Bar')
        # When
        actual = RecordMap.create_record_map(item)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(1, actual._field_count)
        self.assertEqual(item, actual.get_item(0))
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_map_create_with_value(self):
        # Given
        item = Value.create_from('Baz')
        # When
        actual = RecordMap.create_record_map(item)
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(0, actual._field_count)
        self.assertEqual(item, actual.get_item(0))

    def test_record_map_get_item_existing(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('First', '1st')
        second_item = Attr.create_attr('Second', '2nd')
        third_item = Attr.create_attr('Third', '3rd')
        record_map.add(first_item)
        record_map.add(second_item)
        record_map.add(third_item)
        # When
        actual = record_map.get_item(1)
        # Then
        self.assertEqual(second_item, actual)
        self.assertEqual(self.expected_strings.get(self.get_name()), str(actual))

    def test_record_map_get_item_underflow(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('First', '1st')
        second_item = Attr.create_attr('Second', '2nd')
        record_map.add(first_item)
        record_map.add(second_item)
        # When
        actual = record_map.get_item(-1)
        # Then
        self.assertEqual(_Absent._get_absent(), actual)

    def test_record_map_get_item_overflow(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('First', '1st')
        record_map.add(first_item)
        # When
        actual = record_map.get_item(2)
        # Then
        self.assertEqual(_Absent._get_absent(), actual)

    def test_record_map_get_items_empty(self):
        # Given
        record_map = RecordMap.create()
        # When
        actual = record_map.get_items()
        # Then
        self.assertEqual([], actual)

    def test_record_map_get_items_single(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('Hello', 'World')
        record_map.add(first_item)
        # When
        actual = record_map.get_items()
        # Then
        self.assertEqual([first_item], actual)

    def test_record_map_get_items_multiple(self):
        # Given
        record_map = RecordMap.create()
        first_item = Attr.create_attr('Hello', 'World')
        second_item = Attr.create_attr('Bye', 'World')
        record_map.add(first_item)
        record_map.add(second_item)
        # When
        actual = record_map.get_items()
        # Then
        self.assertEqual([first_item, second_item], actual)

    def test_record_map_get_body_empty_without_head(self):
        # Given
        record_map = RecordMap.create()
        # When
        actual = record_map.get_body()
        # Then
        self.assertEqual(_Absent._get_absent(), actual)

    def test_record_map_get_body_empty_with_head(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Hello', 'World'))
        # When
        actual = record_map.get_body()
        # Then
        self.assertEqual(_Absent._get_absent(), actual)

    def test_record_map_get_body_single_value(self):
        # Given
        record_map = RecordMap.create()
        body = Text.create_from('Moo')
        record_map.add(Attr.create_attr('Hello', 'World'))
        record_map.add(body)
        # When
        actual = record_map.get_body()
        # Then
        self.assertEqual(body, actual)

    def test_record_map_get_body_single_object(self):
        # Given
        record_map = RecordMap.create()
        body = Attr.create_attr('Moo', 'Moo')
        record_map.add(Attr.create_attr('Hello', 'World'))
        record_map.add(body)
        # When
        actual = record_map.get_body()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(body, actual.get_item(0))

    def test_record_map_get_body_multiple_(self):
        # Given
        record_map = RecordMap.create()
        first = Attr.create_attr('Moo', 'Moo')
        second = Attr.create_attr('Boo', 'Boo')
        record_map.add(Attr.create_attr('Goodbye', 'World'))
        record_map.add(first)
        record_map.add(second)
        # When
        actual = record_map.get_body()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(first, actual.get_item(0))
        self.assertEqual(second, actual.get_item(1))

    def test_record_immutable_map_add(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = _RecordFlags.IMMUTABLE.value
        item = Attr.create_attr('Moo', 'Moo')
        # When
        with self.assertRaises(TypeError) as error:
            record_map.add(item)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Cannot add item to immutable record!', message)

    def test_record_aliased_map_add(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = _RecordFlags.ALIASED.value
        item = Attr.create_attr('Moo', 'Moo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual([item], record_map._items)
        self.assertEqual(1, record_map._item_count)
        self.assertEqual(1, record_map._field_count)
        self.assertEqual(0, record_map._flags)
        self.assertEqual(None, record_map._fields)

    def test_record_mutable_map_add(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 0
        item = Attr.create_attr('Foo', 'Foo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertEqual([item], record_map._items)
        self.assertTrue(actual_response)
        self.assertEqual(1, record_map._item_count)
        self.assertEqual(1, record_map._field_count)

    def test_record_map_add_mutable_non_field(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 0
        item = Value.create_from('Foo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertEqual([item], record_map._items)
        self.assertTrue(actual_response)
        self.assertEqual(1, record_map._item_count)
        self.assertEqual(0, record_map._field_count)

    def test_record_map_add_mutable_field_with_hashtable_unique(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 0
        first_item = Attr.create_attr('Foo', 'Foo')
        second_item = Attr.create_attr('Moo', 'Moo')

        # When
        record_map.add(first_item)
        record_map.contains_key('Foo')
        record_map.add(second_item)
        # Then
        self.assertEqual([first_item, second_item], record_map._items)
        self.assertEqual(2, record_map._item_count)
        self.assertEqual(2, record_map._field_count)
        self.assertEqual({'Foo': first_item, 'Moo': second_item}, record_map._fields)

    def test_record_map_add_mutable_field_with_hashtable_duplicate(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 0
        first_item = Attr.create_attr('Boo', 'Foo')
        second_item = Attr.create_attr('Boo', 'Moo')
        # When
        record_map.add(first_item)
        record_map.add(second_item)
        record_map.contains_key('Boo')
        # Then
        self.assertEqual([first_item, second_item], record_map._items)
        self.assertEqual(2, record_map._item_count)
        self.assertEqual(2, record_map._field_count)
        self.assertEqual({'Boo': second_item}, record_map._fields)

    def test_record_map_add_aliased_non_field(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 2
        item = Value.create_from('Boo')
        # When
        actual_response = record_map.add(item)
        # Then
        self.assertEqual([item], record_map._items)
        self.assertEqual(True, actual_response)
        self.assertEqual(1, record_map._item_count)
        self.assertEqual(0, record_map._field_count)

    def test_record_map_commit_mutable(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 0
        # When
        actual = record_map.commit()
        # Then
        self.assertEqual(record_map, actual)
        self.assertEqual(_RecordFlags.IMMUTABLE.value, actual._flags)

    def test_record_map_commit_aliased(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 2
        # When
        actual = record_map.commit()
        # Then
        self.assertEqual(record_map, actual)
        self.assertEqual(3, actual._flags)

    def test_record_map_commit_immutable(self):
        # Given
        record_map = RecordMap.create()
        record_map._flags = 1
        # When
        actual = record_map.commit()
        # Then
        self.assertEqual(record_map, actual)
        self.assertEqual(1, actual._flags)

    def test_record_map_contains_key_value(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Moo', 'Boo'))
        # When
        actual = record_map.contains_key(Text.create_from('Moo'))
        # Then
        self.assertTrue(actual)

    def test_record_map_contains_key_string(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Moo', 'Boo'))
        # When
        actual = record_map.contains_key('Foo')
        # Then
        self.assertTrue(actual)

    def test_record_map_contains_key_duplicate(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Foo', 'Boo'))
        # When
        actual = record_map.contains_key('Foo')
        # Then
        self.assertTrue(actual)

    def test_record_map_contains_key_false(self):
        # Given
        record_map = RecordMap.create()
        record_map.add(Attr.create_attr('Foo', 'Poo'))
        record_map.add(Attr.create_attr('Moo', 'Boo'))
        # When
        actual = record_map.contains_key('Poo')
        # Then
        self.assertFalse(actual)

    def test_record_map_contains_key_field_count_empty(self):
        # Given
        record_map = RecordMap.create()
        # When
        actual = record_map.contains_key('Poo')
        # Then
        self.assertFalse(actual)

    def test_record_map_branch_empty(self):
        # Given
        original_record = _Record.create()
        original_record._flags = 0
        # When
        copy_record = original_record._branch()
        # Then
        self.assertEqual(original_record._items, copy_record._items)
        self.assertEqual(original_record._fields, copy_record._fields)
        self.assertEqual(original_record._item_count, copy_record._item_count)
        self.assertEqual(original_record._field_count, copy_record._field_count)
        self.assertEqual(2, original_record._flags)
        self.assertEqual(original_record._flags, copy_record._flags)

    def test_record_map_branch_single(self):
        # Given
        original_record = _Record.create()
        original_record._flags = 0
        original_record.add(Attr.create_attr('Foo', 'Bar'))
        # When
        copy_record = original_record._branch()
        # Then
        self.assertEqual(2, original_record._flags)
        self.assertEqual(original_record._items, copy_record._items)
        self.assertEqual(original_record._fields, copy_record._fields)
        self.assertEqual(1, copy_record._item_count)
        self.assertEqual(1, copy_record._field_count)

    def test_record_map_branch_multiple(self):
        # Given
        original_record = _Record.create()
        original_record._flags = 0
        original_record.add(Attr.create_attr('Foo', 'Bar'))
        original_record.add(Attr.create_attr('Baz', 'Qux'))
        # When
        copy_record = original_record._branch()
        # Then
        self.assertEqual(2, original_record._flags)
        self.assertEqual(original_record._items, copy_record._items)
        self.assertEqual(original_record._fields, copy_record._fields)
        self.assertEqual(2, copy_record._item_count)
        self.assertEqual(2, copy_record._field_count)

    def test_record_map_branch_with_hashtable(self):
        # Given
        original_record = _Record.create()
        original_record._flags = 0
        original_record.add(Attr.create_attr('Bar', 'Baz'))
        original_record.contains_key('Bar')
        # When
        copy_record = original_record._branch()
        # Then
        self.assertIsNotNone(copy_record._fields)
        self.assertEqual(original_record._items, copy_record._items)
        self.assertEqual(original_record._fields, copy_record._fields)
        self.assertEqual(1, copy_record._item_count)
        self.assertEqual(1, copy_record._field_count)

    def test_record_map_branch_mutate_original(self):
        # Given
        original_record = _Record.create()
        original_record.add(Attr.create_attr('K', 'V'))
        original_record.add(Attr.create_attr('A', 'B'))
        # When
        original_record.contains_key('K')
        copy_record = original_record._branch()
        original_record.add(Attr.create_attr(Text.create_from('P'), Text.create_from('V')))
        original_record.contains_key('K')
        # Then
        self.assertEqual(0, original_record._flags)
        self.assertEqual(2, copy_record._flags)
        self.assertNotEqual(original_record._items, copy_record._items)
        self.assertEqual(3, original_record._item_count)
        self.assertEqual(2, copy_record._item_count)
        self.assertEqual(3, original_record._field_count)
        self.assertEqual(2, copy_record._field_count)
        self.assertIsNotNone(original_record._fields)
        self.assertIsNotNone(copy_record._fields)
        self.assertEqual(3, len(original_record._fields))
        self.assertEqual(2, len(copy_record._fields))

    def test_record_map_branch_mutate_copy(self):
        # Given
        original_record = _Record.create()
        original_record.add(Attr.create_attr('F', 'L'))
        original_record.add(Attr.create_attr('M', 'S'))
        # When
        original_record.contains_key('P')
        copy_record = original_record._branch()
        copy_record.add(Attr.create_attr(Text.create_from('P'), Text.create_from('V')))
        copy_record.add(Attr.create_attr(Text.create_from('T'), Text.create_from('B')))
        copy_record.contains_key('H')
        # Then
        self.assertEqual(2, original_record._flags)
        self.assertEqual(0, copy_record._flags)
        self.assertNotEqual(original_record._items, copy_record._items)
        self.assertEqual(2, original_record._item_count)
        self.assertEqual(4, copy_record._item_count)
        self.assertEqual(2, original_record._field_count)
        self.assertEqual(4, copy_record._field_count)
        self.assertIsNotNone(original_record._fields)
        self.assertIsNotNone(copy_record._fields)
        self.assertEqual(2, len(original_record._fields))
        self.assertEqual(4, len(copy_record._fields))

    def test_record_map_view(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('A', 'B'))
        record.add(Attr.create_attr('C', 'D'))
        record.add(Attr.create_attr('E', 'F'))
        record.add(Attr.create_attr('G', 'H'))
        # When
        actual = _RecordMapView(record, 1, 3)
        # Then
        self.assertEqual(2, actual.size)
        self.assertEqual(1, actual._lower)
        self.assertEqual(3, actual._upper)

    def test_record_map_view_get_item_existing(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('Z', 'X'))
        record.add(Attr.create_attr('Y', 'W'))
        record.add(Attr.create_attr('P', 'L'))
        record.add(Attr.create_attr('M', 'N'))
        # When
        actual = _RecordMapView(record, 1, 3)
        # Then
        self.assertEqual('W', actual.get_item(0).value)
        self.assertEqual('L', actual.get_item(1).value)

    def test_record_map_view_get_item_underflow(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('G', 'V'))
        record.add(Attr.create_attr('H', 'B'))
        record.add(Attr.create_attr('J', 'N'))
        record.add(Attr.create_attr('K', 'M'))
        # When
        actual = _RecordMapView(record, 1, 3)
        # Then
        self.assertEqual(_Absent._get_absent(), actual.get_item(-1).value)

    def test_record_map_view_get_item_overflow(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('F', 'V'))
        record.add(Attr.create_attr('Y', 'B'))
        record.add(Attr.create_attr('U', 'N'))
        record.add(Attr.create_attr('I', 'M'))
        # When
        actual = _RecordMapView(record, 1, 3)
        # Then
        self.assertEqual(_Absent._get_absent(), actual.get_item(3).value)

    def test_record_map_view_get_items_none(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('U', 'N'))
        record.add(Attr.create_attr('I', 'M'))
        # When
        actual = _RecordMapView(record, 4, 5)
        # Then
        self.assertEqual([], actual.get_items())

    def test_record_map_view_get_items_single(self):
        # Given
        record = _Record.create()
        first = Attr.create_attr('U', 'N')
        second = Attr.create_attr('A', 'N')
        third = Attr.create_attr('I', 'M')
        record.add(first)
        record.add(second)
        record.add(third)
        # When
        actual = _RecordMapView(record, 1, 2)
        # Then
        self.assertEqual([second], actual.get_items())

    def test_record_map_view_get_items_multiple(self):
        # Given
        record = _Record.create()
        first = Attr.create_attr('A', 'K')
        second = Attr.create_attr('K', 'P')
        third = Attr.create_attr('Q', 'F')
        record.add(first)
        record.add(second)
        record.add(third)
        # When
        actual = _RecordMapView(record, 1, 3)
        # Then
        self.assertEqual([second, third], actual.get_items())

    def test_record_map_view_immutable_add(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('L', 'K'))
        record.add(Attr.create_attr('T', 'P'))
        record.add(Attr.create_attr('R', 'F'))
        new_item = Attr.create_attr('New', 'Item')
        record._flags = 1
        record_map_view = _RecordMapView(record, 1, 3)
        # When
        with self.assertRaises(TypeError) as error:
            record_map_view.add(new_item, 2)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Cannot add item to immutable record!', message)

    def test_record_map_view_out_of_bounds_add(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('T', 'K'))
        record.add(Attr.create_attr('I', 'P'))
        record.add(Attr.create_attr('M', 'F'))
        new_item = Attr.create_attr('Foo', 'Bar')
        record_map_view = _RecordMapView(record, 1, 3)
        # When
        with self.assertRaises(IndexError) as error:
            record_map_view.add(new_item, 5)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Index 5 is out of range!', message)

    def test_record_map_view_add_field(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('T', 'K'))
        record.add(Attr.create_attr('C', 'P'))
        record.add(Attr.create_attr('M', 'F'))
        new_item = Attr.create_attr('Field', 'Item')
        record_map_view = _RecordMapView(record, 1, 3)
        # When
        actual_response = record_map_view.add(new_item, 2)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(3, record_map_view.size)
        self.assertEqual(4, record_map_view._upper)
        self.assertEqual(1, record_map_view._lower)
        self.assertEqual(new_item, record_map_view.get_item(2))
        self.assertEqual(4, record_map_view._record.size)
        self.assertEqual(4, record_map_view._record._field_count)
        self.assertEqual(0, record_map_view._record._flags)
        self.assertEqual(None, record_map_view._record._fields)

    def test_record_map_view_add_value(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('T', 'K'))
        record.add(Attr.create_attr('T', 'P'))
        new_item = Text.create_from('Text')
        record_map_view = _RecordMapView(record, 0, 2)
        # When
        actual_response = record_map_view.add(new_item, 1)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(3, record_map_view.size)
        self.assertEqual(3, record_map_view._upper)
        self.assertEqual(0, record_map_view._lower)
        self.assertEqual(new_item, record_map_view.get_item(1))
        self.assertEqual(3, record_map_view._record.size)
        self.assertEqual(2, record_map_view._record._field_count)
        self.assertEqual(0, record_map_view._record._flags)
        self.assertEqual(None, record_map_view._record._fields)

    def test_record_map_view_add_none_index(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('B', 'H'))
        record.add(Attr.create_attr('E', 'R'))
        new_item = Text.create_from('Foo')
        record_map_view = _RecordMapView(record, 1, 2)
        # When
        actual_response = record_map_view.add(new_item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(2, record_map_view.size)
        self.assertEqual(3, record_map_view._upper)
        self.assertEqual(1, record_map_view._lower)
        self.assertEqual(new_item, record_map_view.get_item(1))

    def test_record_map_view_branch_empty(self):
        # Given
        record = _Record.create()
        record.add(Attr.create_attr('Q', 'H'))
        record.add(Attr.create_attr('W', 'R'))
        record_map_view = _RecordMapView(record, 0, 0)
        # When
        actual = record_map_view._branch()
        # Then
        self.assertEqual(0, actual.size)
        self.assertEqual(0, actual._field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view._record.get_items(), actual._items)
        self.assertEqual([], actual._items)
        self.assertEqual(None, actual._fields)

    def test_record_map_view_branch_single(self):
        # Given
        record = _Record.create()
        first = Attr.create_attr('R', 'H')
        second = Attr.create_attr('Y', 'R')
        record.add(first)
        record.add(second)
        record_map_view = _RecordMapView(record, 0, 1)
        # When
        actual = record_map_view._branch()
        # Then
        self.assertEqual(1, actual.size)
        self.assertEqual(1, actual._field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view._record.get_items(), actual._items)
        self.assertEqual([first], actual._items)
        self.assertEqual(None, actual._fields)

    def test_record_map_view_branch_multiple_fields(self):
        # Given
        record = _Record.create()
        first = Attr.create_attr('T', 'H')
        second = Attr.create_attr('E', 'R')
        third = Attr.create_attr('P', 'Q')
        record.add(first)
        record.add(second)
        record.add(third)
        record_map_view = _RecordMapView(record, 0, 2)
        # When
        actual = record_map_view._branch()
        # Then
        self.assertEqual(2, actual.size)
        self.assertEqual(2, actual._field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view._record.get_items(), actual._items)
        self.assertEqual([first, second], actual._items)
        self.assertEqual(None, actual._fields)

    def test_record_map_view_branch_multiple_values(self):
        # Given
        record = _Record.create()
        second = Text.create_from('Bar')
        third = Text.create_from('Baz')
        fourth = Text.create_from('Qux')
        record.add(Text.create_from('Foo'))
        record.add(second)
        record.add(third)
        record.add(fourth)
        record_map_view = _RecordMapView(record, 1, 4)
        # When
        actual = record_map_view._branch()
        # Then
        self.assertEqual(3, actual.size)
        self.assertEqual(0, actual._field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view._record.get_items(), actual._items)
        self.assertEqual([second, third, fourth], actual._items)
        self.assertEqual(None, actual._fields)

    def test_record_map_view_branch_multiple_values_and_fields(self):
        # Given
        record = _Record.create()
        first = Text.create_from('Bar')
        second = Attr.create_attr('E', 'R')
        third = Text.create_from('Baz')
        fourth = Attr.create_attr('P', 'Q')
        record.add(first)
        record.add(second)
        record.add(third)
        record.add(fourth)
        record_map_view = _RecordMapView(record, 0, 3)
        # When
        actual = record_map_view._branch()
        # Then
        self.assertEqual(3, actual.size)
        self.assertEqual(1, actual._field_count)
        self.assertIsInstance(actual, RecordMap)
        self.assertNotEqual(record_map_view._record.get_items(), actual._items)
        self.assertEqual([first, second, third], actual._items)
        self.assertEqual(None, actual._fields)

    def test_value_builder_add_field_to_empty_record(self):
        # Given
        value_builder = _ValueBuilder()
        item = Attr.create_attr('Foo', 'Boo')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(1, value_builder._record.size)
        self.assertEqual(None, value_builder._value)

    def test_value_builder_add_field_existing_record(self):
        # Given
        value_builder = _ValueBuilder()
        value_builder.add(Attr.create_attr('Baz', 'Qux'))
        item = Attr.create_attr('Foo', 'Boo')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(2, value_builder._record.size)
        self.assertEqual(None, value_builder._value)

    def test_value_builder_add_field_to_empty_record_and_existing_value(self):
        # Given
        value_builder = _ValueBuilder()
        value_builder.add(Text.create_from('Baz'))
        item = Attr.create_attr('Moo', 'Cow')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(2, value_builder._record.size)
        self.assertEqual(None, value_builder._value)
        self.assertIsInstance(value_builder._record.get_item(0), Text)
        self.assertIsInstance(value_builder._record.get_item(1), Attr)

    def test_value_builder_add_value_to_existing_record(self):
        # Given
        value_builder = _ValueBuilder()
        value_builder.add(Attr.create_attr('Moo', 'Cow'))
        value_builder.add(Attr.create_attr('Boo', 'Ghost'))
        item = Text.create_from('Baz')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(3, value_builder._record.size)
        self.assertEqual(None, value_builder._value)
        self.assertIsInstance(value_builder._record.get_item(0), Attr)
        self.assertIsInstance(value_builder._record.get_item(1), Attr)
        self.assertIsInstance(value_builder._record.get_item(2), Text)

    def test_value_builder_add_value_to_empty_record_and_value(self):
        # Given
        value_builder = _ValueBuilder()
        item = Text.create_from('Bar')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(None, value_builder._record)
        self.assertEqual('Bar', value_builder._value.value)

    def test_value_builder_add_value_to_empty_record_exiting_value(self):
        # Given
        value_builder = _ValueBuilder()
        value_builder.add(Text.create_from('Bar'))
        value_builder.add(Text.create_from('Boo'))
        value_builder.add(Text.create_from('Foo'))
        item = Text.create_from('Moo')
        # When
        actual_response = value_builder.add(item)
        # Then
        self.assertTrue(actual_response)
        self.assertEqual(4, value_builder._record.size)
        self.assertEqual(None, value_builder._value)
        self.assertIsInstance(value_builder._record.get_item(0), Text)
        self.assertIsInstance(value_builder._record.get_item(1), Text)

    def test_value_builder_add_not_supported(self):
        # Given
        value_builder = _ValueBuilder()
        item = CustomItem()
        # When
        with self.assertRaises(TypeError) as error:
            value_builder.add(item)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Item of type CustomItem is not supported by the Value Builder', message)

    def test_value_builder_bind_with_record(self):
        # Given
        value_builder = _ValueBuilder()
        attr = Attr.create_attr('Foo', 'Bar')
        value_builder.add(attr)
        # When
        actual = value_builder._bind()
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(1, actual.size)
        self.assertEqual(attr, actual.get_item(0))

    def test_value_builder_bind_with_value(self):
        # Given
        value_builder = _ValueBuilder()
        value = Text.create_from('Foo')
        value_builder.add(value)
        # When
        actual = value_builder._bind()
        # Then
        self.assertEqual(value, actual)

    def test_value_builder_bind_with_empty_record_and_value(self):
        # Given
        value_builder = _ValueBuilder()
        # When
        actual = value_builder._bind()
        # Then
        self.assertEqual(_Absent._get_absent(), actual)

    def test_get_converter_first(self):
        # When
        actual = RecordConverter.get_converter()
        # Then
        self.assertIsInstance(actual, RecordConverter)
        self.assertEqual(RecordConverter.get_converter(), actual)
        self.assertEqual(RecordConverter._converter, actual)

    def test_get_converter_existing(self):
        # Given
        expected = RecordConverter.get_converter()
        # When
        actual = RecordConverter.get_converter()
        # Then
        self.assertIsInstance(actual, RecordConverter)
        self.assertEqual(RecordConverter.get_converter(), actual)
        self.assertEqual(RecordConverter._converter, actual)
        self.assertEqual(expected, actual)

    def test_converter_object_to_record_item(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = Text.create_from('Boo')
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertEqual(obj, actual)

    def test_converter_object_to_record_str(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = 'Moo'
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, Text)
        self.assertEqual('Moo', actual.value)

    def test_converter_object_to_record_float(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = 13.37
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(13.37, actual.value)

    def test_converter_object_to_record_int(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = -1024
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, Num)
        self.assertEqual(-1024, actual.value)

    def test_converter_object_to_record_bool(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = True
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, Bool)
        self.assertEqual(Bool._TRUE, actual)

    def test_converter_object_to_record_dict_single(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = {'a': 5}
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Slot)
        self.assertIsInstance(actual.get_item(0).key, Text)
        self.assertEqual('a', actual.get_item(0).key.value)
        self.assertIsInstance(actual.get_item(0).value, Num)
        self.assertEqual(5, actual.get_item(0).value.value)

    def test_convert_object_to_record_dict_empty(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = {}
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(0, actual.size)

    def test_converter_object_to_record_regular_object(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = MockPerson('John', 20)
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(0), Attr)
        self.assertIsInstance(actual.get_item(1), Slot)
        self.assertIsInstance(actual.get_item(2), Slot)
        self.assertEqual('MockPerson', actual.get_item(0).key.value)
        self.assertEqual('name', actual.get_item(1).key.value)
        self.assertEqual('John', actual.get_item(1).value.value)
        self.assertEqual('age', actual.get_item(2).key.value)
        self.assertEqual(20, actual.get_item(2).value.value)

    def test_converter_object_to_record_nested_object(self):
        # Given
        converter = RecordConverter.get_converter()
        friend = MockPerson('Mike', 23)
        obj = MockPerson('George', 25, friend)
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertIsInstance(actual.get_item(3), Slot)
        self.assertIsInstance(actual.get_item(2), Slot)
        self.assertIsInstance(actual.get_item(1), Slot)
        self.assertIsInstance(actual.get_item(0), Attr)
        self.assertEqual('MockPerson', actual.get_item(0).key.value)
        self.assertEqual('name', actual.get_item(1).key.value)
        self.assertEqual('George', actual.get_item(1).value.value)
        self.assertEqual('age', actual.get_item(2).key.value)
        self.assertEqual(25, actual.get_item(2).value.value)
        self.assertEqual('friend', actual.get_item(3).key.value)
        self.assertIsInstance(actual.get_item(3).value, RecordMap)
        self.assertIsInstance(actual.get_item(3).value.get_item(0), Attr)
        self.assertIsInstance(actual.get_item(3).value.get_item(1), Slot)
        self.assertIsInstance(actual.get_item(3).value.get_item(2), Slot)
        self.assertEqual('MockPerson', actual.get_item(3).value.get_item(0).key.value)
        self.assertEqual('name', actual.get_item(3).value.get_item(1).key.value)
        self.assertEqual('Mike', actual.get_item(3).value.get_item(1).value.value)
        self.assertEqual('age', actual.get_item(3).value.get_item(2).key.value)
        self.assertEqual(23, actual.get_item(3).value.get_item(2).value.value)

    def test_convert_object_to_record_empty(self):
        # Given
        converter = RecordConverter.get_converter()
        obj = None
        # When
        actual = converter.object_to_record(obj)
        # Then
        self.assertIsInstance(actual, RecordMap)
        self.assertEqual(0, actual.size)

    def test_convert_record_to_object_absent(self):
        # Given
        converter = RecordConverter.get_converter()
        record = _Absent._get_absent()
        classes = {}
        strict = False
        # When
        # noinspection PyTypeChecker
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, _Absent)
        self.assertEqual(_Absent._get_absent(), actual)

    def test_convert_record_to_object_text(self):
        # Given
        converter = RecordConverter.get_converter()
        record = Text.create_from("Foo")
        classes = {}
        strict = False
        # When
        # noinspection PyTypeChecker
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, str)
        self.assertEqual("Foo", actual)

    def test_convert_record_to_object_int(self):
        # Given
        converter = RecordConverter.get_converter()
        record = Num.create_from(0)
        classes = {}
        strict = False
        # When
        # noinspection PyTypeChecker
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, int)
        self.assertEqual(0, actual)

    def test_convert_record_to_object_float(self):
        # Given
        converter = RecordConverter.get_converter()
        record = Num.create_from(-75.21)
        classes = {}
        strict = False
        # When
        # noinspection PyTypeChecker
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, float)
        self.assertEqual(-75.21, actual)

    def test_convert_record_to_object_bool(self):
        # Given
        converter = RecordConverter.get_converter()
        record = Bool.create_from(True)
        classes = {}
        strict = False
        # When
        # noinspection PyTypeChecker
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, bool)
        self.assertEqual(True, actual)

    def test_convert_record_to_object_dict_single(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()
        record.add(Slot.create_slot(Text.create_from('Foo'), Text.create_from('Bar')))
        classes = {}
        strict = False
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, dict)
        self.assertEqual('Bar', actual['Foo'])

    def test_convert_record_to_object_dict_multiple(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()
        record.add(Slot.create_slot(Text.create_from('height'), Num.create_from(21)))
        record.add(Slot.create_slot(Text.create_from('width'), Num.create_from(3.14)))
        classes = {}
        strict = False
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, dict)
        self.assertEqual(21, actual['height'])
        self.assertEqual(3.14, actual['width'])

    def test_convert_record_to_object_nested_dicts(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('John')))
        pet_record = RecordMap.create()
        pet_record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Bark')))
        record.add(Slot.create_slot(Attr.create_attr(Text.create_from('pet'), pet_record)))
        classes = {}
        strict = False
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, dict)
        self.assertEqual('John', actual['name'])
        self.assertIsInstance(actual['pet'], dict)
        self.assertEqual('Bark', actual['pet']['name'])

    def test_convert_record_to_object_regular_object(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()
        record.add(Attr.create_attr(Text.create_from('MockPerson'), Value.extant()))
        record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(21)))
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('foo')))
        classes = dict()
        classes['MockPerson'] = MockPerson
        strict = False
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, MockPerson)
        self.assertEqual('foo', actual.name)
        self.assertEqual(21, actual.age)

    def test_convert_record_to_object_nested_objects(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()

        pet_record = RecordMap.create()
        pet_record.add(Attr.create_attr(Text.create_from('MockPet'), Value.extant()))
        pet_record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(33)))
        pet_record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Bark')))

        record.add(Attr.create_attr(Text.create_from('MockPerson'), Value.extant()))
        record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(19)))
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Moo')))
        record.add(Slot.create_slot(Text.create_from('friend'), pet_record))

        classes = dict()
        classes['MockPerson'] = MockPerson
        classes['MockPet'] = MockPet
        strict = False
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, MockPerson)
        self.assertEqual('Moo', actual.name)
        self.assertEqual(19, actual.age)
        self.assertIsInstance(actual.friend, MockPet)
        self.assertEqual('Bark', actual.friend.name)
        self.assertEqual(33, actual.friend.age)

    def test_convert_record_to_object_missing_class_not_strict(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()

        pet_record = RecordMap.create()
        pet_record.add(Attr.create_attr(Text.create_from('MockPet'), Value.extant()))
        pet_record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(11)))
        pet_record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Philip')))

        record.add(Attr.create_attr(Text.create_from('MockPerson'), Value.extant()))
        record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(50)))
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('John')))
        record.add(Slot.create_slot(Text.create_from('friend'), pet_record))

        classes = dict()
        strict = False
        classes['MockPerson'] = MockPerson
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, MockPerson)
        self.assertEqual('John', actual.name)
        self.assertEqual(50, actual.age)
        self.assertEqual('MockPet', type(actual.friend).__name__)
        self.assertEqual('Philip', actual.friend.name)
        self.assertEqual(11, actual.friend.age)

    def test_convert_record_to_object_missing_class_strict(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()

        pet_record = RecordMap.create()
        pet_record.add(Attr.create_attr(Text.create_from('MockPet'), Value.extant()))
        pet_record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(23)))
        pet_record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Michael')))

        record.add(Attr.create_attr(Text.create_from('MockPerson'), Value.extant()))
        record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(28)))
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('Robert')))
        record.add(Slot.create_slot(Text.create_from('friend'), pet_record))

        classes = dict()
        strict = True
        classes['MockPerson'] = MockPerson
        # When
        with self.assertRaises(Exception) as error:
            converter.record_to_object(record, classes, strict)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Missing class for MockPet.', message)

    def test_convert_record_to_object_class_not_matching_not_strict(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()
        record.add(Attr.create_attr(Text.create_from('MockPerson'), Value.extant()))
        record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(39)))
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('baz')))
        classes = dict()
        classes['MockPerson'] = MockCar
        strict = False
        # When
        actual = converter.record_to_object(record, classes, strict)
        # Then
        self.assertIsInstance(actual, MockCar)
        self.assertEqual('baz', actual.name)
        self.assertEqual(39, actual.age)
        self.assertIsNone(actual.model)
        self.assertIsNone(actual.make)
        self.assertIsNone(actual.year)

    def test_convert_record_to_object_class_not_matching_strict(self):
        # Given
        converter = RecordConverter.get_converter()
        record = RecordMap.create()
        record.add(Attr.create_attr(Text.create_from('MockPerson'), Value.extant()))
        record.add(Slot.create_slot(Text.create_from('age'), Num.create_from(11)))
        record.add(Slot.create_slot(Text.create_from('name'), Text.create_from('bar')))
        classes = dict()
        classes['MockPerson'] = MockCar
        strict = True
        with self.assertRaises(Exception) as error:
            converter.record_to_object(record, classes, strict)
        # Then
        message = error.exception.args[0]
        self.assertEqual('Missing attribute age for class MockCar.', message)



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import unittest
from concurrent import futures
from threading import Thread
from unittest.mock import patch

from aiounittest import async_test
from swimai.client._downlinks._downlinks import _ValueDownlinkView, _MapDownlinkView, _EventDownlinkView
from swimai.structures import Text
from test.utils import MockWebsocketConnect, MockWebsocket, MockAsyncFunction, MockScheduleTask, \
    mock_exception_callback, MockRunWithExceptionOnce, MockExceptionOnce
from swimai import SwimClient


class TestSwimClient(unittest.TestCase):

    def setUp(self):
        MockWebsocket.clear()
        MockScheduleTask.clear()
        MockExceptionOnce.clear()
        asyncio.get_event_loop().stop()
        asyncio.get_event_loop().close()

    def test_swim_client_start(self):
        # Given
        client = SwimClient()
        # When
        actual = client.start()
        # Then
        self.assertEqual(actual, client)
        self.assertIsInstance(actual._loop, asyncio.events.AbstractEventLoop)
        self.assertIsInstance(actual._loop_thread, Thread)
        self.assertFalse(actual._loop.is_closed())
        self.assertTrue(actual._loop_thread.is_alive())
        self.assertTrue(actual._has_started)
        client.stop()

    def test_swim_client_stop(self):
        # Given
        client = SwimClient()
        client.start()
        # When
        actual = client.stop()
        # Then
        self.assertEqual(client, actual)
        self.assertIsInstance(actual._loop, asyncio.events.AbstractEventLoop)
        self.assertIsInstance(actual._loop_thread, Thread)
        self.assertTrue(actual._loop.is_closed())
        self.assertFalse(actual._loop_thread.is_alive())
        self.assertFalse(actual._has_started)

    def test_swim_client_with_statement(self):
        # When
        with SwimClient() as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client._loop, asyncio.events.AbstractEventLoop)
            self.assertIsInstance(swim_client._loop_thread, Thread)
            self.assertFalse(swim_client._loop.is_closed())
            self.assertTrue(swim_client._has_started)

        self.assertIsInstance(swim_client._loop, asyncio.events.AbstractEventLoop)
        self.assertIsInstance(swim_client._loop_thread, Thread)
        self.assertTrue(swim_client._loop.is_closed())
        self.assertFalse(swim_client._loop_thread.is_alive())
        self.assertFalse(swim_client._has_started)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_with_statement_exception_no_handler(self, mock_warn_tb, mock_warn):
        # When
        with SwimClient(debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client._loop, asyncio.events.AbstractEventLoop)
            self.assertIsInstance(swim_client._loop_thread, Thread)
            self.assertFalse(swim_client._loop.is_closed())
            self.assertTrue(swim_client._loop_thread.is_alive())

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        mock_warn.assert_called_once()
        mock_warn_tb.assert_called_once()
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0])
        self.assertTrue(swim_client._loop.is_closed())
        self.assertFalse(swim_client._loop_thread.is_alive())

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('os._exit')
    def test_swim_client_with_statement_exception_terminate(self, mock_exit, mock_warn_tb, mock_warn):
        # When
        with SwimClient(terminate_on_exception=True, debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client._loop, asyncio.events.AbstractEventLoop)
            self.assertFalse(swim_client._loop.is_closed())
            self.assertIsInstance(swim_client._loop_thread, Thread)
            self.assertTrue(swim_client._loop_thread.is_alive())

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        self.assertTrue(swim_client._loop.is_closed())
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0])
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        mock_exit.assert_called_once_with(1)
        self.assertFalse(swim_client._loop_thread.is_alive())

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('builtins.print')
    def test_swim_client_with_statement_exception_callback(self, mock_print, mock_warn_tb, mock_warn):
        # Given
        mock_callback = mock_exception_callback
        # When
        with SwimClient(execute_on_exception=mock_callback, debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client._loop, asyncio.events.AbstractEventLoop)
            self.assertTrue(swim_client._loop_thread.is_alive())
            self.assertIsInstance(swim_client, SwimClient)
            self.assertFalse(swim_client._loop.is_closed())
            self.assertIsInstance(swim_client._loop_thread, Thread)

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        self.assertEqual('Mock exception callback', mock_print.call_args_list[0][0][0])
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0])
        self.assertTrue(swim_client._loop.is_closed())
        mock_warn_tb.assert_called_once()
        self.assertFalse(swim_client._loop_thread.is_alive())

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    @patch('os._exit')
    def test_swim_client_with_statement_exception_callback_and_terminate(self, mock_exit, mock_warn_tb, mock_warn):
        # Given
        mock_callback = mock_exception_callback
        # When
        with SwimClient(terminate_on_exception=True, execute_on_exception=mock_callback, debug=True) as swim_client:
            # Then
            self.assertIsInstance(swim_client, SwimClient)
            self.assertIsInstance(swim_client._loop, asyncio.events.AbstractEventLoop)
            self.assertFalse(swim_client._loop.is_closed())
            self.assertTrue(swim_client._loop_thread.is_alive())
            self.assertIsInstance(swim_client._loop_thread, Thread)

            swim_client.task_with_exception = lambda: (_ for _ in ()).throw(Exception,
                                                                            Exception('Mock exception in task'))

            swim_client.task_with_exception()

        mock_exit.assert_called_once_with(1)
        self.assertTrue(swim_client._loop.is_closed())
        self.assertEqual('Mock exception in task', mock_warn.call_args_list[0][0][0])
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        self.assertFalse(swim_client._loop_thread.is_alive())

    @patch('websockets.connect', new_callable=MockWebsocketConnect)
    def test_swim_client_command(self, mock_websocket_connect):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'
        expected = '@command(node:moo,lane:cow)"Hello, World!"'
        with SwimClient() as swim_client:
            # When
            actual = swim_client.command(host_uri, node_uri, lane_uri, Text.create_from('Hello, World!'))
            while not actual.done():
                pass

        # Then
        self.assertIsInstance(actual, futures.Future)
        mock_websocket_connect.assert_called_once_with(host_uri)
        self.assertEqual(expected, MockWebsocket.get_mock_websocket().sent_messages[0])

    def test_swim_client_command_before_open(self):
        # Given
        swim_client = SwimClient()
        # When
        downlink_event = swim_client.downlink_event()
        # Then
        self.assertIsInstance(downlink_event, _EventDownlinkView)
        self.assertEqual(downlink_event._client, swim_client)

    def test_swim_client_downlink_event(self):
        # Given
        with SwimClient() as swim_client:
            # When
            downlink_event = swim_client.downlink_event()

        # Then
        self.assertIsInstance(downlink_event, _EventDownlinkView)
        self.assertEqual(downlink_event._client, swim_client)

    @patch('warnings.warn')
    def test_swim_client_downlink_event_open_before_client_started(self, mock_warn):
        # Given
        swim_client = SwimClient()
        # When
        downlink_event = swim_client.downlink_event()
        downlink_event.open()
        # Then
        self.assertEqual('Cannot execute "_add_downlink_view" before the client has been started!',
                         mock_warn.call_args_list[0][0][0])

    def test_swim_client_downlink_map(self):
        # Given
        with SwimClient() as swim_client:
            # When
            downlink_map = swim_client.downlink_map()

        # Then
        self.assertIsInstance(downlink_map, _MapDownlinkView)
        self.assertEqual(downlink_map._client, swim_client)

    @patch('warnings.warn')
    def test_swim_client_downlink_map_open_before_client_started(self, mock_warn):
        # Given
        swim_client = SwimClient()
        # When
        downlink_map = swim_client.downlink_map()
        downlink_map.open()
        # Then
        self.assertEqual('Cannot execute "_add_downlink_view" before the client has been started!',
                         mock_warn.call_args_list[0][0][0])

    def test_swim_client_downlink_value(self):
        # Given
        with SwimClient() as swim_client:
            # When
            downlink_view = swim_client.downlink_value()

        # Then
        self.assertIsInstance(downlink_view, _ValueDownlinkView)
        self.assertEqual(downlink_view._client, swim_client)

    @patch('warnings.warn')
    def test_swim_client_downlink_value_open_before_client_started(self, mock_warn):
        # Given
        swim_client = SwimClient()
        # When
        downlink_value = swim_client.downlink_value()
        downlink_value.open()
        # Then
        self.assertEqual('Cannot execute "_add_downlink_view" before the client has been started!',
                         mock_warn.call_args_list[0][0][0])

    @patch('swimai.client._connections._ConnectionPool._add_downlink_view', new_callable=MockAsyncFunction)
    @async_test
    async def test_swim_client_add_downlink_view(self, mock_add_downlink):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'

        MockWebsocket.get_mock_websocket().messages_to_send.append('@synced(node:"moo",lane:"cow")')

        with SwimClient() as swim_client:
            downlink_view = swim_client.downlink_value()
            downlink_view._host_uri = host_uri
            downlink_view._node_uri = node_uri
            downlink_view._lane_uri = lane_uri
            # When
            await swim_client._add_downlink_view(downlink_view)

        # Then
        mock_add_downlink.assert_called_once_with(downlink_view)

    @patch('swimai.client._connections._ConnectionPool._add_downlink_view', new_callable=MockAsyncFunction)
    @patch('swimai.client._connections._ConnectionPool._remove_downlink_view', new_callable=MockAsyncFunction)
    @async_test
    async def test_swim_client_remove_downlink_view(self, mock_remove_downlink, mock_add_downlink):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'

        MockWebsocket.get_mock_websocket().messages_to_send.append('@synced(node:"moo",lane:"cow")')

        with SwimClient() as swim_client:
            downlink_view = swim_client.downlink_value()
            downlink_view._host_uri = host_uri
            downlink_view._node_uri = node_uri
            downlink_view._lane_uri = lane_uri
            await swim_client._add_downlink_view(downlink_view)
            # When
            await swim_client._remove_downlink_view(downlink_view)

        # Then
        mock_add_downlink.assert_called_once_with(downlink_view)
        mock_remove_downlink.assert_called_once_with(downlink_view)

    @patch('swimai.client._connections._ConnectionPool._get_connection', new_callable=MockAsyncFunction)
    @async_test
    async def test_swim_client_get_connection(self, mock_get_connection):
        # Given
        host_uri = 'ws://localhost:9001'
        node_uri = 'moo'
        lane_uri = 'cow'

        MockWebsocket.get_mock_websocket().messages_to_send.append('@synced(node:"moo",lane:"cow")')

        with SwimClient() as swim_client:
            downlink_view = swim_client.downlink_value()
            downlink_view._host_uri = host_uri
            downlink_view._node_uri = node_uri
            downlink_view._lane_uri = lane_uri
            # When
            await swim_client._get_connection(host_uri)

        # Then
        mock_get_connection.assert_called_once_with(host_uri)

    @async_test
    async def test_swim_client_test_schedule_task(self):
        #  Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient() as swim_client:
            # When
            actual = swim_client._schedule_task(mock_task.async_execute, 'foo')
            while not actual.done():
                pass

        # Then
        self.assertEqual(1, mock_task.call_count)
        self.assertEqual('foo', mock_task.message)
        self.assertIsInstance(actual, futures.Future)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_test_schedule_task_exception(self, mock_print_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        original_run = asyncio.run_coroutine_threadsafe

        with patch('asyncio.run_coroutine_threadsafe', new_callable=MockRunWithExceptionOnce) as mock_run:
            mock_exception_once = MockExceptionOnce.get_mock_exception_once()
            mock_exception_once.actual_function = original_run
            mock_run.side_effect = mock_exception_once.side_effect

            # When
            with SwimClient() as swim_client:
                actual = swim_client._schedule_task(mock_task.sync_execute, 'foo')

            self.assertEqual('Mock exception', mock_warn.call_args_list[0][0][0])
            mock_print_tb.assert_not_called()
            self.assertIsNone(actual)
            self.assertEqual(2, mock_run.call_count)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_test_schedule_task_exception_debug(self, mock_print_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        original_run = asyncio.run_coroutine_threadsafe

        with patch('asyncio.run_coroutine_threadsafe', new_callable=MockRunWithExceptionOnce) as mock_run:
            mock_exception_once = MockExceptionOnce.get_mock_exception_once()
            mock_exception_once.actual_function = original_run
            mock_run.side_effect = mock_exception_once.side_effect

            # When
            with SwimClient(debug=True) as swim_client:
                actual = swim_client._schedule_task(mock_task.sync_execute, 'foo')

            self.assertEqual('Mock exception', mock_warn.call_args_list[0][0][0])
            mock_print_tb.assert_called_once()
            self.assertIsNone(actual)

    @patch('warnings.warn')
    @patch('traceback.print_tb')
    def test_swim_client_test_schedule_task_that_raises_exception(self, mock_warn_tb, mock_warn):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient(debug=True) as swim_client:
            # When
            actual = swim_client._schedule_task(mock_task.async_exception_execute, 'foo')
            while not actual.done():
                pass

        # Then
        mock_warn_tb.assert_called_once()
        mock_warn.assert_called_once()
        self.assertEqual('Mock async execute exception', mock_warn.call_args_list[0][0][0])
        self.assertEqual(1, mock_task.call_count)
        self.assertEqual('foo', mock_task.message)
        self.assertIsInstance(actual, futures.Future)

    def test_swim_client_test_schedule_task_that_is_cancelled(self):
        # Given
        mock_task = MockScheduleTask.get_mock_schedule_task()
        with SwimClient() as swim_client:
            # When
            actual = swim_client._schedule_task(mock_task.async_infinite_cancel_execute)
        # Then
        self.assertIsInstance(actual, futures.Future)
        self.assertTrue(actual.cancelled())



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from typing import Any
from unittest.mock import MagicMock
from swimai.client._connections import _ConnectionStatus
from swimai.structures._structs import _Item


class CustomString:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other

    def __str__(self):
        return f'CustomString({self.value})'


class CustomItem(_Item):
    @property
    def key(self) -> 'Any':
        return 'MockKey'

    @property
    def value(self) -> 'Any':
        return 'MockVale'


class MockAsyncFunction(MagicMock):
    async def __call__(self, *args, **kwargs):
        return super(MockAsyncFunction, self).__call__(*args, **kwargs)


class MockWebsocketConnect(MagicMock):

    async def __call__(self, *args, **kwargs):
        return super(MockWebsocketConnect, self).__call__(*args, **kwargs)

    @property
    def return_value(self):
        return MockWebsocket.get_mock_websocket()

    @staticmethod
    def set_raise_exception(state):
        MockWebsocket.get_mock_websocket().raise_exception = state


class MockWebsocketConnectException(MagicMock):

    async def __call__(self, *args, **kwargs):
        return super(MockWebsocketConnectException, self).__call__(*args, **kwargs)

    @property
    def return_value(self):
        raise Exception('Mock_websocket_connect_exception')


class MockDownlink:
    instance = None

    def __init__(self):
        self.all_messages_sent = asyncio.Event()
        self.call_counts_left = 0

    async def receive_message(self):

        self.call_counts_left = self.call_counts_left - 1

        if self.call_counts_left == 0:
            self.all_messages_sent.set()

    @staticmethod
    def get_mock_downlink():
        if MockDownlink.instance is None:
            MockDownlink.instance = MockDownlink()

        return MockDownlink.instance

    @staticmethod
    def clear():
        MockDownlink.instance = None


class MockReceiveMessage(MagicMock):

    async def __call__(self, *args, **kwargs):
        await MockDownlink.get_mock_downlink().receive_message()
        return super(MockReceiveMessage, self).__call__(*args, **kwargs)

    @staticmethod
    def set_call_count(call_count):
        MockDownlink.get_mock_downlink().call_counts_left = call_count

    @staticmethod
    def all_messages_has_been_sent():
        return MockDownlink.get_mock_downlink().all_messages_sent


class MockWebsocket:
    instance = None

    def __init__(self):
        self.connection = None
        self.closed = False
        self.sent_messages = list()
        self.messages_to_send = list()
        self.raise_exception = False
        self.custom_recv_func = None

    @staticmethod
    def get_mock_websocket():
        if MockWebsocket.instance is None:
            MockWebsocket.instance = MockWebsocket()

        return MockWebsocket.instance

    @staticmethod
    def clear():
        MockWebsocket.instance = None

    async def close(self):
        self.closed = True

    async def send(self, message):
        self.sent_messages.append(message)

    async def recv(self):

        if self.custom_recv_func is None:

            if self.raise_exception:
                raise Exception('WebSocket Exception!')

            message = self.messages_to_send.pop()

            if len(self.messages_to_send) == 0:
                self.connection.status = _ConnectionStatus.CLOSED

            return message

        else:
            return await self.custom_recv_func()


class ReceiveLoop:

    def __init__(self):
        self.call_count = 0

    async def recv_loop(self):
        while True:
            self.call_count = self.call_count + 1
            await asyncio.sleep(1)


class MockConnection:
    instance = None

    def __init__(self):
        self.owner = None
        self.messages_sent = list()
        self.messages_to_receive = list()

    @staticmethod
    def get_mock_connection():
        if MockConnection.instance is None:
            MockConnection.instance = MockConnection()

        return MockConnection.instance

    @staticmethod
    def clear():
        MockConnection.instance = None

    async def _wait_for_messages(self):
        while True:
            if len(self.messages_to_receive) > 0 and self.owner is not None:
                message = self.messages_to_receive.pop()
                await self.owner._receive_message(message)

            await asyncio.sleep(1)

    async def _send_message(self, message):
        self.messages_sent.append(message)


def mock_did_set_confirmation():
    print(1)
    pass


async def mock_did_set_callback(old, new):
    str(old)
    str(new)
    pass


async def mock_on_event_callback(event):
    str(event)
    pass


async def mock_did_update_callback(key, old, new):
    str(key)
    str(old)
    str(new)
    pass


async def mock_did_remove_callback(key, old):
    str(key)
    str(old)
    pass


def mock_exception_callback():
    print('Mock exception callback')


class MockExceptionOnce:
    instance = None

    def __init__(self):
        self.calls = 0
        self.actual_function = None

    @staticmethod
    def get_mock_exception_once():
        if MockExceptionOnce.instance is None:
            MockExceptionOnce.instance = MockExceptionOnce()

        return MockExceptionOnce.instance

    @staticmethod
    def clear():
        MockExceptionOnce.instance = None

    def side_effect(self, *args, **kwargs):
        if self.calls == 0:
            self.calls = self.calls + 1
            self.exception()
        else:
            return self.normal(*args, **kwargs)

    @staticmethod
    def exception():
        raise Exception('Mock exception')

    def normal(self, *args, **kwargs):
        return self.actual_function(*args, **kwargs)


class MockRunWithExceptionOnce(MagicMock):

    def __call__(self, *args, **kwargs):
        return super(MockRunWithExceptionOnce, self).__call__(*args, **kwargs)


class MockScheduleTask:
    instance = None

    def __init__(self):
        self.message = None
        self.call_count = 0

    @staticmethod
    async def async_execute(message):
        MockScheduleTask.instance.message = message
        MockScheduleTask.instance.call_count = MockScheduleTask.instance.call_count + 1

    @staticmethod
    def sync_execute(message):
        MockScheduleTask.instance.message = message
        MockScheduleTask.instance.call_count = MockScheduleTask.instance.call_count + 1

    @staticmethod
    async def async_exception_execute(message):
        MockScheduleTask.instance.message = message
        MockScheduleTask.instance.call_count = MockScheduleTask.instance.call_count + 1
        raise Exception('Mock async execute exception')

    @staticmethod
    async def async_infinite_cancel_execute():
        while True:
            await asyncio.sleep(1)

    @staticmethod
    def get_mock_schedule_task():
        if MockScheduleTask.instance is None:
            MockScheduleTask.instance = MockScheduleTask()

        return MockScheduleTask.instance

    @staticmethod
    def clear():
        MockScheduleTask.instance = None


class MockClass:

    def __init__(self):
        pass


class MockPerson:

    def __init__(self, name=None, age=None, friend=None):
        self.name = name
        self.age = age
        self.friend = friend


class MockPet:
    def __init__(self, name=None, age=None):
        self.name = name
        self.age = age


class MockCar:

    def __init__(self, make=None, model=None, year=None):
        self.make = make
        self.model = model
        self.year = year


def mock_func():
    return 'mock_func_response'


async def mock_coro():
    return 'mock_coro_response'


class MockExecuteOnException:
    instance = None

    def __init__(self):
        self.called = False

    def __call__(self, *args, **kwargs):
        self.called = True

    @staticmethod
    def get_mock_execute_on_exception():
        if MockExecuteOnException.instance is None:
            MockExecuteOnException.instance = MockExecuteOnException()

        return MockExecuteOnException.instance

    @staticmethod
    def clear():
        MockExecuteOnException.instance = None


class NewScope:
    class MockPerson:
        def __init__(self, name='Uriel'):
            self.name = name


class MockNoDefaultConstructor:

    def __init__(self, name):
        self.name = name


class MockModel:
    pass


class MockDownlinkManager:

    def __init__(self):
        self.called = 0
        self.event = None
        self.did_set_new = None
        self.did_set_old = None
        self.update_key = None
        self.update_value_new = None
        self.update_value_old = None
        self.remove_key = None
        self.remove_old_value = None
        self.strict = False
        self.registered_classes = dict()

    async def _subscribers_on_event(self, event):
        self.called = self.called + 1
        self.event = event

    async def _subscribers_did_set(self, did_set_new, did_set_old):
        self.called = self.called + 1
        self.did_set_new = did_set_new
        self.did_set_old = did_set_old

    async def _subscribers_did_update(self, update_key, update_value_new, update_value_old):
        self.called = self.called + 1
        self.update_key = update_key
        self.update_value_new = update_value_new
        self.update_value_old = update_value_old

    async def _subscribers_did_remove(self, remove_key, remove_old_value):
        self.called = self.called + 1
        self.remove_key = remove_key
        self.remove_old_value = remove_old_value


class MockEventCallback:

    def __init__(self):
        self.called = False
        self.event = None

    async def execute(self, event):
        self.called = True
        self.event = event


class MockDidSetCallback:
    def __init__(self):
        self.called = False
        self.new_value = None
        self.old_value = None

    async def execute(self, new_value, old_value):
        self.called = True
        self.new_value = new_value
        self.old_value = old_value


class MockDidUpdateCallback:
    def __init__(self):
        self.called = False
        self.key = None
        self.new_value = None
        self.old_value = None

    async def execute(self, key, new_value, old_value):
        self.called = True
        self.key = key
        self.new_value = new_value
        self.old_value = old_value


class MockDidRemoveCallback:
    def __init__(self):
        self.called = False
        self.key = None
        self.value = None

    async def execute(self, key, value):
        self.called = True
        self.key = key
        self.value = value



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from typing import Callable
from urllib.parse import urlparse, ParseResult


def after_started(function: 'Callable') -> 'Callable':
    """
    Decorator that allows for a given client method to be executed only after the client has been started.
    The client should have an `has_started` flag set to True.

    :param function:        - Decorated client method.
    :return:                - The return value of the decorated client method.
    """

    def wrapper(*args, **kwargs):
        if args[0]._has_started:
            return function(*args, **kwargs)
        else:
            try:
                raise Exception(f'Cannot execute "{args[1].__name__}" before the client has been started!')
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                args[0]._handle_exception(exc_value, exc_traceback)

    return wrapper


class _URI:

    @staticmethod
    def _normalise_warp_scheme(uri: str) -> str:
        """
        Normalise all different representations of the WARP scheme to a websocket connection.

        :param uri:             - URI to normalise.
        :return:                - URI with `ws` scheme.
        """
        uri = urlparse(uri)
        if _URI._has_valid_scheme(uri):
            uri = uri._replace(scheme='ws')
            return uri.geturl()
        else:
            raise TypeError(f'Invalid scheme "{uri.scheme}" for Warp URI!')

    @staticmethod
    def _has_valid_scheme(uri: ParseResult) -> bool:
        """
        Check if a URI has a WARP compatible scheme.

        :param uri:             - URI to check.
        :return:                - True if the given URI has a scheme that is WARP compatible. False otherwise.
        """
        scheme = uri.scheme
        return scheme == 'ws' or scheme == 'warp'



#  Copyright 2015-2020 SWIM.AI inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Union, Any

from ._utils import _ReconUtils, _InputMessage, _OutputMessage
from swimai.structures import Text, Bool, Attr, Value, Slot, Num, RecordMap
from swimai.structures._structs import _Item, _Record, _ValueBuilder


class _ReconParser:

    @staticmethod
    async def _create_ident(value: str) -> 'Value':
        if value == 'true':
            return Bool.create_from(True)
        elif value == 'false':
            return Bool.create_from(False)

        return Text.create_from(value)

    @staticmethod
    async def _create_attr(key: Any, value: Any = Value.extant()) -> 'Attr':
        return Attr.create_attr(key, value)

    @staticmethod
    async def _create_record_builder() -> 'RecordMap':
        return _Record.create()

    @staticmethod
    async def _create_value_builder() -> '_ValueBuilder':
        return _ValueBuilder()

    @staticmethod
    async def _create_slot(key: Any, value: Any = None) -> 'Slot':
        return Slot.create_slot(key, value)

    @staticmethod
    async def _create_number(value: Union[float, int]) -> 'Num':
        return Num.create_from(value)

    async def _parse_block_string(self, recon_string: str) -> 'Value':
        message = await _InputMessage._create(recon_string)
        return await self._parse_block_expression(message)

    async def _parse_block_expression(self, message: '_InputMessage') -> 'Value':
        return await self._parse_attr_expression(message)

    async def _parse_attr_expression(self, message: '_InputMessage', builder: Union[RecordMap, _ValueBuilder] = None,
                                     field_output: 'Value' = None, value_output: 'Value' = None) -> 'Value':
        return await _AttrExpressionParser._parse(message=message, parser=self, builder=builder,
                                                  field_output=field_output, value_output=value_output)

    async def _parse_attr(self, message: '_InputMessage', key_output: 'Value' = None,
                          value_output: 'Value' = None) -> 'Attr':

        return await _AttrParser._parse(message=message, parser=self, key_output=key_output, value_output=value_output)

    async def _parse_string(self, message: '_InputMessage', output: '_OutputMessage' = None) -> 'Text':
        return await _StringParser._parse(message=message, parser=self, output=output)

    async def _parse_number(self, message: '_InputMessage', value_output: int = None, sign_output: int = 1) -> 'Num':
        return await _NumberParser._parse(message=message, parser=self, value_output=value_output,
                                          sign_output=sign_output)

    async def _parse_literal(self, message: '_InputMessage',
                             builder: Union[RecordMap, _ValueBuilder] = None) -> 'Value':
        return await _LiteralParser._parse(message=message, parser=self, builder=builder)

    async def _parse_record(self, message: '_InputMessage', builder: Union[RecordMap, _ValueBuilder] = None,
                            key_output: 'Value' = None, value_output: 'Value' = None) -> 'Value':
        return await _RecordParser._parse(message=message, parser=self, builder=builder, key_output=key_output,
                                          value_output=value_output)

    async def _parse_ident(self, message: '_InputMessage', output: '_OutputMessage' = None) -> 'Value':
        return await _IdentParser._parse(message=message, parser=self, output=output)


class _AbstractParser(ABC):
    @staticmethod
    @abstractmethod
    async def _parse() -> '_Item':
        """
        Parse a message into its Item object representation.

        :return:                - Item object.
        """
        raise NotImplementedError


class _RecordParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = _InputMessage(), parser: '_ReconParser' = None,
                     builder: Union[RecordMap, _ValueBuilder] = None, key_output: 'Value' = None,
                     value_output: 'Value' = None) -> 'Value':

        if builder is None:
            builder = await parser._create_record_builder()

        char = message._head

        if char == '{' or char == '[':
            message._step()

        await message._skip_spaces(message)

        if key_output is None:
            key_output = await parser._parse_block_expression(message)

        await message._skip_spaces(message)

        if message._is_cont:
            if message._head == ':':
                message._step()

                if value_output is None:
                    value_output = await parser._parse_block_expression(message)

            builder.add(await parser._create_slot(key_output, value_output))
            key_output = None

            char = message._head

            if char == ',' or char == ';':
                message._step()
                await _RecordParser._parse(message, parser, builder)

            elif char == '}' or char == ']':
                message._step()

        if key_output:
            builder.add(await parser._create_slot(key_output, value_output))

        return builder._bind()


class _AttrExpressionParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = _InputMessage(), parser: '_ReconParser' = None,
                     builder: Union[RecordMap, _ValueBuilder] = None, field_output: 'Value' = None,
                     value_output: 'Value' = None) -> 'Value':

        await message._skip_spaces(message)

        char = message._head

        if char == '@':

            if field_output is None:
                field_output = await parser._parse_attr(message)

            if builder is None:
                builder = await parser._create_record_builder()

            builder.add(field_output)
            await _AttrExpressionParser._parse(message, parser, builder)
            return builder._bind()

        elif await _ReconUtils._is_ident_start_char(char) or char == '"' or await _ReconUtils._is_digit(
                char) or char == '-':

            if value_output is None:
                value_output = await parser._parse_literal(message)

            if builder is None:
                builder = await parser._create_value_builder()

            builder.add(value_output)
            await _AttrExpressionParser._parse(message, parser, builder)
            return builder._bind()

        elif char == '{' or char == '[':
            return await parser._parse_literal(message, builder)


class _AttrParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = _InputMessage(), parser: '_ReconParser' = None,
                     key_output: 'Value' = None, value_output: 'Value' = None) -> 'Attr':

        await message._skip_spaces(message)

        char = message._head

        if char == '@':
            message._step()

            if key_output is None:
                key_output = await parser._parse_ident(message)

            if message._head == '(' and message._is_cont:
                message._step()
            else:
                return await parser._create_attr(key_output)

            if value_output is None:
                value_output = await parser._parse_record(message)
                message._step()
                return await parser._create_attr(key_output, value_output)

        if key_output is None or value_output is None:
            raise TypeError(f'Attribute starting at position {message.index} is invalid!\nMessage: {message._value}')
        else:
            return await parser._create_attr(key_output, value_output)


class _IdentParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = _InputMessage(), parser: '_ReconParser' = None,
                     output: '_OutputMessage' = None) -> 'Value':

        await message._skip_spaces(message)

        char = message._head

        if await _ReconUtils._is_ident_start_char(char):
            if output is None:
                output = await _OutputMessage._create()

            await output._append(char)
            char = message._step()

            while await _ReconUtils._is_ident_char(char):
                await output._append(char)
                char = message._step()

        if output is not None:
            return await parser._create_ident(output._value)
        else:
            raise TypeError(f'Identifier starting at position {message.index} is invalid!\nMessage: {message._value}')


class _StringParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = None, parser: '_ReconParser' = None,
                     output: '_OutputMessage' = None) -> 'Text':

        await message._skip_spaces(message)

        char = message._head

        if char == '"':

            if output is None:
                output = await _OutputMessage._create()

            char = message._step()

            while char != '"' and message._is_cont:
                await output._append(char)
                char = message._step()

            message._step()

        if output is None:
            output = _OutputMessage()

        return Text.create_from(output._value)


class _NumberParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = None, parser: '_ReconParser' = None, value_output: int = None,
                     sign_output: int = 1) -> Num:

        await message._skip_spaces(message)

        char = message._head

        if char == '-':
            sign_output = -1
            char = message._step()

        while char == '0':
            char = message._step()

        if '1' <= char <= '9':
            value_output = sign_output * int(char)
            char = message._step()

            while message._is_cont and await _ReconUtils._is_digit(char):
                value_output = 10 * value_output + sign_output * int(char)
                char = message._step()

        if message._is_cont and char == '.':
            return await _DecimalParser._parse(message, parser, value_output, sign_output)
        else:
            if value_output is None:
                value_output = 0

            return await parser._create_number(value_output)


class _DecimalParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = _InputMessage(), parser: '_ReconParser' = None,
                     value_output: int = None,
                     sign_output: int = 0) -> 'Num':

        await message._skip_spaces(message)

        builder = await _OutputMessage._create('')

        if sign_output < 0 and value_output is None:
            await builder._append('-0')
        else:
            if value_output is None:
                value_output = 0

            await builder._append(value_output)

        char = message._head

        if char == '.':
            await builder._append('.')
            message._step()

            if message._is_cont:
                char = message._head

                if await _ReconUtils._is_digit(char):
                    await builder._append(char)
                    char = message._step()

                    while message._is_cont and await _ReconUtils._is_digit(char):
                        await builder._append(char)
                        char = message._step()

        return await parser._create_number(float(builder._message))


class _LiteralParser(_AbstractParser):

    @staticmethod
    async def _parse(message: '_InputMessage' = _InputMessage(), parser: '_ReconParser' = None,
                     builder: Union[RecordMap, _ValueBuilder] = None, value_output: int = None) -> Value:

        await message._skip_spaces(message)

        char = message._head

        if char == '{' or char == '[':

            if builder is None:
                builder = await parser._create_record_builder()

            await parser._parse_record(message, builder)

        elif await _ReconUtils._is_ident_start_char(char):
            value_output = await parser._parse_ident(message)
        elif char == '"':
            value_output = await parser._parse_string(message)
        elif char == '-' or await _ReconUtils._is_digit(char):
            value_output = await parser._parse_number(message)

        if builder is None:
            builder = await parser._create_value_builder()

        if value_output is not None:
            builder.add(value_output)

        return builder._bind()
